(**************************************************************************)
(*                                                                        *)
(*  This file is part of WP plug-in of Frama-C.                           *)
(*                                                                        *)
(*  Copyright (C) 2007-2011                                               *)
(*    CEA (Commissariat a l'energie atomique                              *)
(*         et aux energies alternatives)                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Db_types
open Cil_types

type scope =
  | SC_Global
  | SC_Function_in    (* Just before the pre-state *)
  | SC_Function_frame (* Just after the introduction of formals *)
  | SC_Function_out   (* Post-state *)
  | SC_Block_in
  | SC_Block_out

type assigns_method =
  | NoAssigns
  | NormalAssigns
  | EffectAssigns
 
module type Export =
sig
  type pred
  val export_goal : Format.formatter -> string -> pred -> unit
  val export_decl : Format.formatter -> pred Formula.declaration -> unit
end

module type Splitter =
sig
  type pred
  val split : assigns_method -> pred -> pred Bag.t
end

(**
 * This is what is really needed to propagate something through the CFG.
 * Usually, the propagated thing should be a predicate,
 * but it can be more sophisticated like lists of predicates,
 * or maybe a structure to keep hypotheses and goals separated.
 * Moreover, proof obligations may also need to be handeled.
 **)
module type S = sig

  type t_prop
  val pretty : Format.formatter -> t_prop -> unit
  val merge : t_prop -> t_prop -> t_prop
  val empty : t_prop

  type t_env
  val new_env : kernel_function -> t_env

  val add_axiom : WpAnnot.prop_id -> string -> logic_label list -> predicate named -> unit
  val add_hyp  : t_env -> WpAnnot.prop_id -> predicate named -> t_prop -> t_prop
  val add_goal : t_env -> WpAnnot.prop_id -> predicate named -> t_prop -> t_prop

  (** [add_as_hyp env h p] add h as hypothesis in p (h => p) *)
  val add_as_hyp  : t_env -> t_prop -> t_prop -> t_prop

  (** [add_as_goal env g p] add g as goal to p (g /\ p) *)
  val add_as_goal : t_env -> t_prop -> t_prop -> t_prop

  val add_assigns : 
    t_env -> WpAnnot.prop_id -> 
    logic_label -> 
    WpAnnot.a_kind -> identified_term assigns -> t_prop -> t_prop

  (** [use_assigns env hid kind assgn goal] performs the havoc on the
      goal. [hid] should be [None] iff [assgn] is [WritesAny], and tied to
      the corresponding identified_property otherwise.
   *)
  val use_assigns : 
    t_env -> WpAnnot.prop_id option ->
    WpAnnot.a_kind -> identified_term assigns -> t_prop -> t_prop

  val assigns_method : unit -> assigns_method

  val label :  t_env -> Clabels.c_label -> t_prop -> t_prop
  val assign : t_env -> lval -> Cil_types.exp -> t_prop -> t_prop
  val return : t_env -> exp option -> t_prop -> t_prop
  val test : t_env -> exp -> t_prop -> t_prop -> t_prop
  val switch : t_env -> exp -> (exp list * t_prop) list -> t_prop -> t_prop

  (* -------------------------------------------------------------------------- *)
  (* --- Call Rules                                                         --- *)
  (* -------------------------------------------------------------------------- *)

  val call_goal_precond : t_env -> stmt -> 
    kernel_function -> exp list ->
    pre: (WpAnnot.prop_id * predicate named) list ->
    t_prop -> t_prop

  val call : t_env -> stmt -> 
    lval option -> kernel_function -> exp list ->
    pre:     (WpAnnot.prop_id * predicate named) list ->
    post:    (WpAnnot.prop_id * predicate named) list ->
    pexit:   (WpAnnot.prop_id * predicate named) list ->
    assigns: identified_term assigns ->
    p_post: t_prop -> 
    p_exit: t_prop ->
    t_prop

  (* -------------------------------------------------------------------------- *)
  (* --- SCOPING RULES                                                      --- *)
  (* -------------------------------------------------------------------------- *)

  val scope : t_env -> varinfo list -> scope -> t_prop -> t_prop

  (** build [p => alpha(p)] for functional dependencies verification. *)
  val build_prop_of_from : t_env -> WpAnnot.t_id_pred list -> t_prop -> t_prop

  val close : t_env -> t_prop -> t_prop

end
