(**************************************************************************)
(*                                                                        *)
(*  This file is part of WP plug-in of Frama-C.                           *)
(*                                                                        *)
(*  Copyright (C) 2007-2011                                               *)
(*    CEA (Commissariat a l'energie atomique                              *)
(*         et aux energies alternatives)                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* -------------------------------------------------------------------------- *)
(* --- Prover Implementation against Task API                             --- *)
(* -------------------------------------------------------------------------- *)

open Task
open Wpo
type verdict = Valid | Invalid | Unknown

(* -------------------------------------------------------------------------- *)
(* --- File Preparation Utilities                                         --- *)
(* -------------------------------------------------------------------------- *)

let cat files cout =
  let buffer = String.create 2048 in
  List.iter
    (fun f ->
       let cin = open_in f in
       try
	 Command.bincopy buffer cin cout ;
	 close_in cin
       with e ->
	 close_in cin ; raise e)
    files

let export file preludes pp =
  let cout = open_out file in
  try
    cat preludes cout ;
    flush cout ;
    let fmt = Format.formatter_of_out_channel cout in
    pp fmt ;
    Format.pp_print_newline fmt () ;
    Format.pp_print_flush fmt () ;
    close_out cout ;
  with err ->
    close_out cout ;
    raise err

module Tmp =
struct

  let tmp = ref 0
  let file base ext =
    if Wp_parameters.debug_atleast 1
    then
      begin
	let dout = Wp_parameters.get_output () in
	incr tmp ; Printf.sprintf "%s/wptmp%d%s" dout !tmp ext
      end
    else Extlib.temp_file_cleanup_at_exit base ext

end

(*managment of the file PO_log.txt generated into the output_dir *)
module Log =
struct

  type t = {
    logout : out_channel ;
    logfmt : Format.formatter;
    mutable logtmp : string list;
    mutable logcmd : (string * string array) option ;
  }

  let create w =
    let file = Wpo.file_for_log ~gid:w.po_gid in
    let outc = open_out file in
    let fmt = Format.formatter_of_out_channel outc in
    {
      logout = outc ;
      logfmt = fmt ;
      logtmp = [] ;
      logcmd = None ;
    }

  let pp_command fmt cmd args =
    begin
      Format.fprintf fmt "@[<hov 4>%s" cmd ;
      Array.iter
	(fun arg -> Format.fprintf fmt "@ %s" arg)
	args ;
      Format.fprintf fmt "@]@." ;
    end

  let pp_current fmt t =
    match t.logcmd with
      | None -> ()
      | Some (cmd,args) -> pp_command fmt cmd args

  let pp_status fmt =
    function
      | Task.Result res ->
	  if res <> 0 then Format.fprintf fmt "Exit [%d]@." res
      | Task.Canceled ->
	  Format.fprintf fmt "Timeout@."
      | Task.Failed exn ->
	  Format.fprintf fmt "Failed \"%s\"@." (Printexc.to_string exn)

  let command t cmd args =
    begin
      t.logcmd <- Some (cmd,args) ;
      pp_command t.logfmt cmd args ;
    end

  let log_status t st = Format.fprintf t.logfmt "Run %a" pp_status st

  let is_error ?(status=[0]) = function
    | Task.Result r -> not (List.mem r status)
    | Task.Canceled -> false
    | Task.Failed _ -> true

  let output t ?status (stdout:Buffer.t) st =
    let msg = Buffer.contents stdout in
    begin
      log_status t st ;
      Format.pp_print_string t.logfmt msg ;
      Format.pp_print_newline t.logfmt () ;
      if is_error ?status st then
	begin
	  let cname = match t.logcmd with None -> "<?>" | Some (cmd,_) -> cmd in
	  if Wp_parameters.debug_atleast 1 then
	    begin
	      Wp_parameters.error "command '%s' failed." cname ;
	      Log.print_on_output "%a%s" pp_current t msg ;
	    end
	  else
	    Wp_parameters.error "command '%s' failed (-wp-debug 1 for details)" cname
	end
    end

  (* Add tempory files*)
  let add_tmp_file t file = 
    Wp_parameters.debug ~level:1 "adding tmp file '%s'" file ;
    t.logtmp <- file::t.logtmp

  (* removes tempory generated files *)
  let clean_up_tmp_file files =
    List.iter
      (fun file ->
	 try
	   Unix.unlink file;
	   Wp_parameters.debug ~level:1 "remove tmp file '%s'" file
	 with _ -> (*TODO: Unix error only *)
	   Wp_parameters.debug ~level:1 "tmp file '%s' removed yet" file)
      files

  let close_out t =
    try
      Format.pp_print_flush t.logfmt () ;
      Pervasives.close_out t.logout ;
    with _ ->
      Wp_parameters.failure "Can not close log file"

  let close t =
    begin
      close_out t ;
      t.logcmd <- None ;
      clean_up_tmp_file t.logtmp ;
      t.logtmp <- [] ;
    end

end

(* -------------------------------------------------------------------------- *)
(* --- Why & Other SMT                                                    --- *)
(* -------------------------------------------------------------------------- *)

module Why =
struct

  let goal log prover w =
    let file = Tmp.file w.po_gid ".why" in
    export file [
      file_for_model ~model:w.po_model prover ;
      file_for_env  ~env:w.po_env prover ;
      file_for_goal ~gid:w.po_gid prover ;
    ] (fun _fmt -> ()) ;
    Log.add_tmp_file log file ;
    file

end

module SMT =
struct

  type t = {
    ext : string ; (* file extension *)
    why : string array ; (* why options *)
    wdp : string array ; (* why-dp options *)
  }

  let yices = {
    ext = "smt" ;
    why = [| "--smtlib" |] ;
    wdp = [| "-smt-solver" ; "yices" |] ;
  }

  let cvc3 = {
    ext = "smt" ;
    why = [| "--smtlib" |] ;
    wdp = [| "-smt-solver" ; "cvc3" |] ;
  }

  let z3 = {
    ext = "z3.smt" ;
    why = [| "--z3" |] ;
    wdp = [| "-smt-solver" ; "z3" |] ;
  }

  let simplify = {
    ext = "sx" ;
    why = [| "--simplify" |] ;
    wdp = [| |] ;
  }

  let altergo = {
    ext = "why" ;
    why = [| "--alt-ergo" |] ;
    wdp = [| |] ;
  }

  let zenon = {
    ext = "znn" ;
    why = [| "--zenon" |] ;
    wdp = [| |] ;
  }

  let of_name dp =
    match String.lowercase dp with
      | "yices" -> yices
      | "cvc3" -> cvc3
      | "z3" -> z3
      | "simplify" -> simplify
      | "alt-ergo" -> altergo
      | "zenon" -> zenon
      | _ -> Wp_parameters.abort "Unknown prover '%s'" dp

  let translate log smt w =
    let goal = Why.goal log L_why w in
    let base = Filename.chop_suffix goal ".why" in
    let file = Printf.sprintf "%s_why.%s" base smt.ext in
    let args = Array.append smt.why [| goal |] in
    let stdout = Buffer.create 512 in
    Log.command log "why" args ;
    Task.command ~stdout "why" args
    >>? Log.output log stdout
    >>= fun res ->
      if res = 0 then Task.return file
      else Task.failed "Why exit %d" res

  let whydp log smt file =
    let timeout = Wp_parameters.Timeout.get () in
    let stdout = Buffer.create 80 in
    let args = Array.append smt.wdp [| "-batch" ; file |] in
    Log.command log "why-dp" args ;
    Task.command ~timeout ~stdout "why-dp" args
    >>? Log.output log ~status:[0;1;2;3;4] stdout
    >>= fun res ->
      if res = 0 then Task.return Valid
      else
	if (1<= res && res <= 4) then Task.return Unknown
	else Task.failed "Why-dp exit %d" res

  let prove dp log w =
    let smt = of_name dp in
    translate log smt w >>= whydp log smt

  let check log w =
    let timeout = Wp_parameters.Timeout.get () in
    let stdout = Buffer.create 80 in
    let goal = Why.goal log L_why w in
    let args = [| "--type-only" ; goal|] in
    Log.command log "why" args;
    Task.command ~timeout ~stdout "why" args
    >>? Log.output log stdout
    >>= fun res ->
      if res = 0 then Task.return Valid else Task.return Invalid

end

(* -------------------------------------------------------------------------- *)
(* --- Prover Alt-Ergo                                                    --- *)
(* -------------------------------------------------------------------------- *)

module AltErgo =
struct

  let valid = Str.regexp "\\bValid\\b"

  let prove arrays log w =
    let timeout = Wp_parameters.Timeout.get () in
    let stdout = Buffer.create 80 in
    let goal = Why.goal log (if arrays then L_ergo92 else L_ergo91) w in
    let args = if arrays then [| "-arrays" ; goal |] else [| goal |] in
    Log.command log "alt-ergo" args ;
    Task.command ~timeout ~stdout "alt-ergo" args
    >>? Log.output log stdout
    >>= fun s ->
      if s=0 then
	let response = Buffer.contents stdout in
	try
	  ignore (Str.search_forward valid response 0) ;
	  Task.return Valid
	with Not_found ->
	  Task.return Unknown
      else
	Task.failed "Alt-Ergo exit %d" s

  let check arrays log w =
     let timeout = Wp_parameters.Timeout.get () in
    let stdout = Buffer.create 80 in
    let goal = Why.goal log (if arrays then L_ergo92 else L_ergo91) w in
    let args = if arrays then [| "-type-only";"-arrays" ; goal |] else [| goal |] in
    Log.command log "alt-ergo" args ;
    Task.command ~timeout ~stdout "alt-ergo" args
    >>? Log.output log stdout
    >>= fun s ->
      if (s = 0) then Task.return Valid
      else Task.return Invalid

end


(* -------------------------------------------------------------------------- *)
(* --- Prover Coq                                                         --- *)
(* -------------------------------------------------------------------------- *)


module Coq =
struct

  let compiled : (string,unit Task.task) Hashtbl.t = Hashtbl.create 83
  let revert_on_error key = function
    | Task.Result _ -> ()
    | _ -> Hashtbl.remove compiled key
  let once key cc =
    try Hashtbl.find compiled key
    with Not_found ->
      let t = Task.nop >>= cc >>? revert_on_error key in
      Hashtbl.add compiled key t ; t

  let result_compile file r =
    if r<>0
    then Task.failed "Compilation of '%s' failed" file
    else Task.return ()

  let require_wp log =
    once "wp"
      (fun () ->
	 let denv = Wp_parameters.get_output () in
	 let dshare = Wp_parameters.get_share() in
	 let file = Format.sprintf "%s/wp.v" dshare in
	 let fcoq = Format.sprintf "%s/wp.v" denv in
	 Log.add_tmp_file log fcoq  ;
	 Command.copy file fcoq ;
	 let args = [| "-noglob" ; fcoq |] in
	 let stdout = Buffer.create 512 in
	 let timeout = Wp_parameters.Timeout.get() in
	 Log.command log "coqc" args ;
	 Task.command ~timeout ~stdout "coqc" args
	 >>? Log.output log stdout
	 >>= result_compile file)

  let require_model log w =
    let model = w.po_model in
    once model
      (fun () ->
	 let file = Wpo.file_for_model ~model L_coq in
	 let fcoq = Wpo.coqc_for_model ~model in
	 Log.add_tmp_file log fcoq ;
	 Command.copy file fcoq ;
	 let denv = Wp_parameters.get_output () in
	 let args = [| "-noglob" ; "-I" ; denv ; fcoq |] in
	 let stdout = Buffer.create 512 in
	 let timeout =  Wp_parameters.Timeout.get() in
	 Log.command log "coqc" args ;
	 Task.command ~timeout ~stdout "coqc" args
	 >>? Log.output log stdout
	 >>= result_compile fcoq)

  let require_env log w =
    let env = w.po_env in
    once env
      (fun () ->
	 let denv = Wp_parameters.get_output () in
	 let file = file_for_env env L_coq in
	 let args =  [| "-noglob"; "-I" ; denv ; file |] in
	 let stdout = Buffer.create 512 in
	 let timeout =  Wp_parameters.Timeout.get() in
	 Log.command log "coqc" args ;
	 Task.command ~timeout ~stdout "coqc" args
	 >>? Log.output log stdout
	 >>= result_compile file)

  let default =
    "  (** FILL PROOF HERE.**)\n  \
       intros; repeat(split; intros); auto;\n  \
       try contradiction; auto; eauto ; try omega.\n"

  let goal log w script =
    let gid = w.po_gid in
    let env = w.po_env in
    let model = w.po_model in
    let file = Tmp.file gid ".v" in
    let coq_file = Wpo.file_for_goal ~gid L_coq in
    Log.add_tmp_file log file ;
    Log.add_tmp_file log coq_file ;
    Command.pp_to_file file
      (fun fmt ->
	 Format.fprintf fmt "Require Import Reals.@\n";
	 Format.fprintf fmt "Require Import wp.@\n";
	 Format.fprintf fmt "Require Import %s.@\n" (Wpo.coq_for_model ~model);
	 Format.fprintf fmt "Require Import %s.@\n" (Wpo.coq_for_env ~env);
	 Command.pp_from_file fmt coq_file;
	 begin
	   match script with
	     | Some script ->
		 Format.fprintf fmt "Proof.@\n%sQed.@\n@." script ;
	     | None -> (* This case is for check*)
		  Format.fprintf fmt "Proof.@\nAdmitted.@\n@." ;
	 end
      );
    Task.return file

  let coqc log w script =
    goal log w script >>=
      fun fgoal ->
        let fgoal = Filename.chop_suffix fgoal ".v" in
	let denv = Wp_parameters.get_output() in
	let timeout = Wp_parameters.Timeout.get() in
	let stdout = Buffer.create 512 in
	let args = [|  "-noglob";"-I" ; denv ; "-compile"; fgoal |] in
	Log.command log "coqtop" args ;
	Task.command ~timeout ~stdout "coqtop" args
	>>? Log.output ~status:[0;1] log stdout

  let script_for w =
    Some
      ( match Proof.find_script_for_goal w.po_gid with
	  | None -> default
	  | Some script -> script )

  let prove log w =
   require_wp log
    >>= fun _ -> require_model log w
      >>= fun _ -> require_env log w
	>>= fun _ -> coqc log w (script_for w)
	  >>= Task.call (fun r -> if r=0 then Valid else Unknown)

  let check log w =
    require_wp log
    >>= fun _ -> require_model log w
      >>= fun _ -> require_env log w
	>>= fun _ -> coqc log w None
	  >>= Task.call (fun r -> if r=0 then Valid else Invalid)


end

module Coqide =
struct

  let coqidelock = Task.mutex ()

  let run_coqide log w fgoal =
    let denv = Wp_parameters.get_output() in
    let gid = w.po_gid in
    let coqide () =
      let head  = Wpo.file_for_head gid in
      let env   = file_for_env w.po_env L_coq in
      let model = coqc_for_model ~model:w.po_model in
      let dshare = Wp_parameters.get_share() in
      let wp_share = Format.sprintf "%s/wp.v" dshare in 
      let args =
	match Wp_parameters.Script.get() with
	  | fscript when Sys.file_exists fscript ->
	      [|  "-I" ; denv ; fgoal; head; env; model; wp_share; fscript |]
	  | _ ->
	      [|  "-I" ; denv ; fgoal; head; env; model; wp_share |]
      in
      Log.command log "coqide" args ;
      Task.command "coqide" args
      >>? Log.log_status log
    in
    Task.sync coqidelock coqide >>=
      fun s ->
	if s <> 0 then Task.return Unknown
	else
	  match Proof.parse_coqproof fgoal with
	    | None ->
		Wp_parameters.feedback "No proof found" ;
		Task.return Unknown
	    | Some script ->
		Wp_parameters.feedback "Check proof" ;
		Coq.coqc log w (Some script) >>=
		  Task.call
		    (fun s ->
		       Proof.add_script gid [] script ;
		       if s=0 then Valid else Unknown)

  let prove log w =
    Coq.prove log w >>= function
      | Unknown | Invalid ->
	  Coq.goal log w (Coq.script_for w) >>=
	    run_coqide log w
      | Valid -> Task.return Valid

end

(* -------------------------------------------------------------------------- *)
(* --- Main Dispatcher                                                    --- *)
(* -------------------------------------------------------------------------- *)

let task_for interactive = function
  | AltErgo -> AltErgo.prove (Wp_parameters.Arrays.get())
  | Why dp -> SMT.prove dp
  | Coq -> if interactive then Coqide.prove else Coq.prove
  | WP -> fun _ _ -> Task.return Unknown

let result = function
  | Task.Result Valid -> Wpo.Valid
  | Task.Result Invalid -> Wpo.Invalid
  | Task.Result Unknown -> Wpo.Unknown
  | Task.Canceled -> Wpo.Timeout
  | Task.Failed exn -> Wpo.Failed (Printexc.to_string exn)

let prove
    ?(callin=fun _ _ -> ())
    ?(callout=fun _ _ _ -> ())
    wpo ~interactive prover
    =
  Task.todo
    begin
      fun () ->
	callin wpo prover ;
	let log = Log.create wpo in
	task_for interactive prover log wpo
	>>! fun s ->
	  let r = result s in
	  Log.close log ;
	  Wpo.set_result wpo prover r ;
	  callout wpo prover r
    end

let check_by = function
  | L_ergo91 -> AltErgo.check false
  | L_ergo92 -> AltErgo.check true
  | L_why -> SMT.check
  | L_coq -> Coq.check

let check
  ?(callout=fun _ _ _ -> ())
    wpo lang
    =
  Task.todo
    begin
      fun () ->
	let log = Log.create wpo in
	check_by lang log wpo
	  >>! fun s ->
	    let r = result s in
	    Log.close log ;
	    callout wpo lang r
    end

let server = ref None

let server () =
  match !server with
    | Some s ->
	let procs = Wp_parameters.Procs.get () in
	Task.set_procs s procs ; s
    | None ->
	let procs = Wp_parameters.Procs.get () in
	let s = Task.server ~procs () in
	Task.on_server_stop s Proof.savescripts ;
	server := Some s ; s
