(**************************************************************************)
(*                                                                        *)
(*  This file is part of WP plug-in of Frama-C.                           *)
(*                                                                        *)
(*  Copyright (C) 2007-2011                                               *)
(*    CEA (Commissariat a l'energie atomique                              *)
(*         et aux energies alternatives)                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* This file groups functions that extract some annotations
 * and associates them with CFG edges. *)

open Cil_types
open Db_types

(* -------------------------------------------------------------------------- *)
(* --- Helpers                                                            --- *)
(* -------------------------------------------------------------------------- *)

let is_main_entry_point kf =
  try
    let main_kf, library = Globals.entry_point () in
      (not library) && (Kernel_function.equal kf main_kf)
  with Globals.No_such_entry_point _ -> false

let get_called_kf fct = match fct.enode with
  | Lval (Var vkf, NoOffset) -> Some (Globals.Functions.get vkf)
  | _ -> None

(* -------------------------------------------------------------------------- *)
(* --- Global Status                                                      --- *)
(* -------------------------------------------------------------------------- *)

let rte_find rte_st kf = let (_,_,status) = !rte_st () in status kf
let rte_precond_status    = rte_find Db.RteGen.get_precond_status
let rte_signedOv_status   = rte_find Db.RteGen.get_signedOv_status
let rte_divMod_status     = rte_find Db.RteGen.get_divMod_status
let rte_downCast_status   = rte_find Db.RteGen.get_downCast_status
let rte_memAccess_status  = rte_find Db.RteGen.get_memAccess_status
let rte_unsignedOv_status = rte_find Db.RteGen.get_unsignedOv_status

let rte_wp =
  [
    "valid pointer dereferencing" , rte_memAccess_status , "-rte-mem" ;
    "division by zero" , rte_divMod_status , "-rte-div" ;
    "signed overflow" , rte_signedOv_status , "-rte-signed" ;
    "unsigned overflow" , rte_unsignedOv_status , "-rte-unsigned-ov" ;
  ]
    
let rte_generated kf =
  List.for_all (fun (_,rte,_) -> rte kf) rte_wp

let missing_rte kf =
  List.map
    (fun (name,_,_) -> name)
    (List.filter (fun (_,rte,_) -> not (rte kf)) rte_wp)

let compute_rte_for kf =
  begin
    Parameters.Dynamic.Bool.set "-rte" true ;
    List.iter (fun (_,_,opt) -> Parameters.Dynamic.Bool.set opt true) rte_wp ;
    !Db.RteGen.annotate_kf kf ;
  end

let ip_complete f = Property.ip_complete_of_spec f Kglobal f.spec

let ip_disjoint f = Property.ip_disjoint_of_spec f Kglobal f.spec

(* -------------------------------------------------------------------------- *)
(* --- Contract for function with code                                    --- *)
(* -------------------------------------------------------------------------- *)

let ip_contract f =
  let ip_behaviors =
    List.map 
      (fun bhv -> Property.ip_of_behavior f Kglobal bhv) f.spec.spec_behavior
  in
  let ip_rte =
    if not (rte_generated f) then
      Wp_parameters.warning ~current:false ~once:true
	"Missing RTE guards" ;
    [ (*TODO: these dependencies should be put on the emitter at Wpo level *) ]
  in
  ip_behaviors @ ip_complete f @ ip_disjoint f @ ip_rte

(* -------------------------------------------------------------------------- *)
(* --- Contract for function without code                                 --- *)
(* -------------------------------------------------------------------------- *)

let ip_external f =
  ip_complete f @ ip_disjoint f

(* -------------------------------------------------------------------------- *)
(* --- Contract for functions                                             --- *)
(* -------------------------------------------------------------------------- *)

module FunctionContracts = Wprop.Indexed2
  (Dashtbl.Default_key_marshaler(Kernel_function))
  (Dashtbl.Default_key_marshaler(Datatype.String))
  (struct
     type key = kernel_function * string
     let size = 81
     let kind = `Correctness
     let dependencies = [ Kernel_function.self ]
     let state_name (kf,model) =
       Printf.sprintf 
	 "Function '%s' is consistent with %s"
	 (Kernel_function.get_name kf) model

     let state_proxy (kf,_) _ip =
       match kf.fundec with
	 | Definition _ -> 
	     Some ("WP Function" , ip_contract kf)
	 | Declaration _ ->
	     Wp_parameters.warning ~current:false ~once:true
	       "Assuming contracts of external functions" ;
	     Some ("WP External Function" , ip_external kf)
       
   end)

(* -------------------------------------------------------------------------- *)
(* --- Local Preconditions for functions                                  --- *)
(* -------------------------------------------------------------------------- *)

module PreCondAt = Wprop.Indexed2
  (Dashtbl.Default_key_marshaler(Property))
  (Dashtbl.Default_key_marshaler(Cil_datatype.Stmt))
  (struct
     type key = Property.t * stmt
     let size = 81
     let kind = `Correctness
     let dependencies = [ Kernel_function.self ]
     let state_name (pid,stmt) =
       Pretty_utils.sfprintf
	 "Precondition @[%a@] at stmt %d" 
	 Property.pretty pid stmt.sid 
     let state_proxy _key _ip = None
   end)

module PreCondProxy = Wprop.Updater
  (struct
     let name = "WP Call Preconditions"
     let kind = `Correctness
     let dependencies = [Ast.self]
   end)

module PreCondProxyGenerated = 
  State_builder.Hashtbl(Property.Hashtbl)(Datatype.Unit)
    (struct
       let name = "WP Call Preconditions Generated"
       let dependencies = [Ast.self]
       let kind = `Internal
       let size = 97
     end)

let pre_cond_id = function
  | Property.IPCodeAnnot(_,_,p) -> p.annot_id
  | Property.IPPredicate(_,_,_,p) -> p.ip_id
  | property -> Wp_parameters.fatal "No precondition id for @[%a@]" Property.pretty property

let setup_precondition_proxy called_kf precondition =
  if not (PreCondProxyGenerated.mem precondition) then
    begin
      let called_preconditions =
	List.map 
	  (fun (_,stmt) -> PreCondAt.property precondition stmt)
	  (Kernel_function.find_syntactic_callsites called_kf)
      in
      PreCondProxy.set_checked precondition called_preconditions Cil_types.True ;
      PreCondProxyGenerated.add precondition ()
    end

(* -------------------------------------------------------------------------- *)
(* --- Loop Consolidation Status                                          --- *)
(* -------------------------------------------------------------------------- *)

module LoopInvariants = Wprop.Indexed
  (Dashtbl.Default_key_marshaler(Cil_datatype.Stmt))
  (struct
     type key = stmt
     let size = 81
     let kind = `Correctness
     let dependencies = [ Ast.self ]
     let state_name stmt =
       Printf.sprintf "Loop Invariants (sid:%d)" stmt.sid
     let state_proxy _s _p = None
   end)

module LoopConsolidation = Wprop.Updater
  (struct
     let name = "WP Loop Invariants Consolidation"
     let kind = `Correctness
     let dependencies = [Ast.self]
   end)

(* -------------------------------------------------------------------------- *)
(* --- Status of Unreachable Annotations                                  --- *)
(* -------------------------------------------------------------------------- *)

module WPUnreachable = Wprop.Updater
  (struct
     let name = "Unreachable Annotations"
     let kind = `Correctness
     let dependencies = [Ast.self]
   end)

(*----------------------------------------------------------------------------*)
(* Property identification                                                    *)
(*----------------------------------------------------------------------------*)

(* Beside the property identification, it can be found in different contexts
 * depending on which part of the computation is involved.
 * For instance, properties on loops are split in 2 parts : establishment and
 * preservation.
 *)

type prop_kind =
  | PKProp        (* normal property *)
  | PKEstablished (* computation related to a loop property before the loop. *)
  | PKPreserved   (* computation related to a loop property inside the loop. *)
  | PKPropLoop    (* loop property used as hypothesis inside a loop. *)
  | PKVarDecr     (* computation related to the decreasing of a variant in a loop *)
  | PKVarPos      (* computation related to a loop variant being positive *)
  | PKAFctOut     (* computation related to the function assigns on normal termination *)
  | PKAFctExit    (* computation related to the function assigns on exit termination *)
  | PKPre of kernel_function * stmt * int (* precondition for function at stmt, pid of the require *)
  
type prop_id = {
  p_kind : prop_kind ;
  p_prop : Property.t ;
  p_part : (int * int) option ;
}

let pretty = ref (fun _ _ -> ())

(* Find back the property identification whatever the context *)
let property_of_id p = p.p_prop

let pid_for_function kf ~model = 
  let prop = FunctionContracts.property kf model in
  { p_kind=PKProp ; p_prop=prop ; p_part=None }

(* proofs *)

module PropSet = Property.Set

type proof = {
  target : Property.t ;
  proved : proofpart array ;
  induction : stmt option ;
  mutable dependencies : PropSet.t ;
  mutable invariants : PropSet.t ;
} and proofpart =
  | Noproof
  | Complete
  | Parts of Bitvector.t

type induction = stmt option

let is_assigns p =
  match p.p_prop with
    | Property.IPAssigns _ -> true
    | _ -> false

let is_requires p =
  match p.p_prop with
    | Property.IPPredicate (Property.PKRequires _,_,_,_) -> true
    | _ -> false

let get_stmt = function
  | Property.IPCodeAnnot(_,stmt,_) -> Some stmt
  | Property.IPAssigns(_,Kstmt stmt,_,_) -> Some stmt
  | _ -> None
      
let get_induction p =
  match p.p_kind with
    | PKProp|PKAFctOut|PKAFctExit|PKPre _ -> None
    | PKEstablished|PKVarDecr|PKVarPos|PKPreserved|PKPropLoop -> get_stmt p.p_prop

let target p = p.target
let dependencies p = 
  let hs = PropSet.elements p.dependencies in
  match p.induction with
    | None -> hs
    | Some stmt -> LoopInvariants.property stmt :: hs

let subproofs = function
  | PKProp | PKPre _ | PKPropLoop -> 1
  | PKEstablished | PKPreserved
  | PKVarDecr | PKVarPos
  | PKAFctExit | PKAFctOut -> 2

let subproof_idx = function
  | PKProp | PKPre _ | PKPropLoop -> 0 (* 1/1 *)
  | PKPreserved  -> 0 (* 1/2 *)
  | PKEstablished-> 1 (* 2/2 *)
  | PKVarDecr    -> 0 (* 1/2 *)
  | PKVarPos     -> 1 (* 2/2 *)
  | PKAFctOut    -> 0 (* 1/2 *)
  | PKAFctExit   -> 1 (* 2/2 *)

let create_proof p =
  let n = subproofs p.p_kind in
  {
    target = p.p_prop ;
    proved = Array.create n Noproof ;
    dependencies = PropSet.empty ;
    invariants = PropSet.empty ;
    induction = get_induction p ;
  }

let add_proof pf p hs =
  begin
    if not (Property.equal (property_of_id p) pf.target)
    then Wp_parameters.fatal "Partial proof inconsistency" ;
    List.iter
      (fun h ->
	 if not (is_requires h) then
	   let hind = get_induction h in
	   let ip = property_of_id h in
	   match pf.induction , hind with
	     | Some s0 , Some s when s0.sid = s.sid ->
		 pf.invariants <- PropSet.add ip pf.invariants
	     | _ ->
		 pf.dependencies <- PropSet.add ip pf.dependencies
      ) hs ;
    let k = subproof_idx p.p_kind in
    match p.p_part with
      | None -> pf.proved.(k) <- Complete
      | Some(p,n) ->
	  match pf.proved.(k) with
	    | Complete -> ()
	    | Noproof -> 
		let bv = Bitvector.create n in
		Bitvector.set_range bv 0 (p-1) ;
		Bitvector.set_range bv (p+1) (n-1) ;
		pf.proved.(k) <- Parts bv
	    | Parts bv -> 
		Bitvector.clear bv p ;
		if Bitvector.is_empty bv 
		then pf.proved.(k) <- Complete
  end

let is_composed pf =
  Array.length pf.proved > 1
    
let is_proved pf =
  try Array.iter (fun r -> if r<>Complete then raise Exit) pf.proved ; true
  with Exit -> false

let consolidate_loop proof =
  match proof.induction with
    | None -> ()
    | Some stmt ->
	try
	  (* Check for consolidation *)
	  PropSet.iter 
	    (fun ip ->
	       match fst (Properties_status.strongest ip) with
		 | Checked{valid=True} -> ()
		 | _ -> raise Exit
	    ) proof.invariants ;
	  (* Consolidated ! *)
	  let iploop = LoopInvariants.property stmt in
	  LoopConsolidation.set_checked iploop [] True  
	with Exit -> () (* not yet consolidated *)

(* prop_id constructors *)

let mk_prop kind prop = { p_kind=kind ; p_prop=prop ; p_part=None }

let mk_annot_id kf stmt ca = Property.ip_of_code_annot_single kf stmt ca

let mk_code_annot_id kf s ca = mk_prop PKProp  (mk_annot_id kf s ca)
let mk_establish_id  kf s ca = mk_prop PKEstablished (mk_annot_id kf s ca)
let mk_preserve_id   kf s ca = mk_prop PKPreserved (mk_annot_id kf s ca)
let mk_inv_hyp_id    kf s ca = mk_prop PKPropLoop  (mk_annot_id kf s ca)
let mk_var_decr_id   kf s ca = mk_prop PKVarDecr (mk_annot_id kf s ca)
let mk_var_pos_id    kf s ca = mk_prop PKVarPos  (mk_annot_id kf s ca)

let mk_disj_bhv_id (kf,ki,disj)  = 
  mk_prop PKProp (Property.ip_of_disjoint kf ki disj)
let mk_compl_bhv_id (kf,ki,comp) = 
  mk_prop PKProp (Property.ip_of_complete kf ki comp)
let mk_decrease_id (kf, s, x)  = 
  mk_prop PKProp (Property.ip_of_decreases kf s x)
let mk_axiom_id name  = mk_prop PKProp (Property.ip_axiom name)
let mk_stmt_assigns_id (kf,ki,b,a) =
  mk_prop PKProp 
    (Extlib.the (Property.ip_of_assigns kf ki b (Writes a)))
let mk_loop_assigns_id (kf,ki,ca,a) =
  mk_prop PKPropLoop
    (Extlib.the (Property.ip_of_assigns kf ki ca (Writes a)))

let mk_fout_assigns_id (kf,ki,b,a) = 
  mk_prop PKAFctOut 
    (Extlib.the (Property.ip_of_assigns kf ki b (Writes a)))

let mk_exit_assigns_id (kf,ki,b,a) = 
  mk_prop PKAFctExit 
    (Extlib.the (Property.ip_of_assigns kf ki b (Writes a)))

let mk_pre_id kf ki b p =
  mk_prop PKProp (Property.ip_of_requires kf ki b p)

let mk_stmt_post_id kf s b p =
  mk_prop PKProp (Property.ip_of_ensures kf (Kstmt s) b p)

let mk_fct_post_id kf b p =
  mk_prop PKProp (Property.ip_of_ensures kf Kglobal b p)

(* prop-id splitter *)

let split job pid goals =
  let n = Bag.length goals in
  if n <= 1 then Bag.iter (job pid) goals else
    let k = ref 0 in
    Bag.iter
      (fun g ->
	 let pid_k = { pid with p_part = Some(!k,n) } in
	 incr k ; job pid_k g)
      goals

(* utilities *)

let pp_names fmt l =  match l with [] -> ()
  | _ ->
      Format.fprintf fmt "_%a" (Wp_error.pp_string_list ~empty:"" ~sep:"_") l

let code_annot_names ca = match ca.annot_content with
  | AAssert (_, named_pred) | AInvariant (_,_,named_pred) -> named_pred.name
  | AVariant (term, _) -> term.term_name
  | _ -> [] (* TODO : add some more names ? *)

(* This is used to give the name of the property that the user can give
 * to select it from the command line (-wp-prop option) *)
let user_prop_names p = match property_of_id p with
    | Property.IPPredicate (_,_,_,idp) -> idp.ip_name
    | Property.IPCodeAnnot (_,_, ca) -> code_annot_names ca
    | Property.IPComplete (_, _, lb) ->
        let name =
          Pretty_utils.sfprintf  "complete_behaviors%a" pp_names lb
        in [name]
    | Property.IPDisjoint (_, _, lb) ->
        let name = Pretty_utils.sfprintf  "disjoint_behaviors%a" pp_names lb
        in [name]
    | Property.IPAssigns (_, _, _, l) ->
      List.fold_left 
        (fun acc (t,_) -> t.it_content.term_name @ acc) ["assigns"] l
    | Property.IPFrom _ -> ["from"] (* TODO: steal term names from assigns? *)
    | Property.IPAxiom _
    | Property.IPDecrease _ -> (*TODO: add more names ? *) []
    | Property.IPBehavior _
    | Property.IPBlob _ -> []

let string_of_termination_kind = function
    Normal -> "post"
  | Exits -> "exits"
  | Breaks -> "breaks"
  | Continues -> "continues"
  | Returns -> "returns"

let predicate_kind_txt pk ki = match pk, ki with
  | Property.PKRequires _, Kglobal -> "pre"
  | Property.PKRequires _, Kstmt _ -> "stmt_pre"
  | Property.PKAssumes _, _ -> "assume"
  | Property.PKEnsures (_, tk), Kglobal ->
      string_of_termination_kind tk
  | Property.PKEnsures (_, tk), Kstmt _ ->
      "stmt_" ^ (string_of_termination_kind tk)
  | Property.PKTerminates, _ -> "terminates"

let id_prop_txt p = match p with
    | Property.IPPredicate (pk,_,ki,idp) ->
        Pretty_utils.sfprintf "%s_%d%a"
         (predicate_kind_txt pk ki)  idp.ip_id pp_names idp.ip_name
    | Property.IPCodeAnnot (_,_, ca) ->
        let name = match ca.annot_content with
          | AAssert _ -> "assert"
          | AInvariant _ -> "loop_inv"
          | APragma _ -> "code_annot"
          | _ -> assert false
        in Pretty_utils.sfprintf "%s_%d%a" name ca .annot_id
          pp_names (code_annot_names ca)
    | Property.IPComplete (_, _, lb) ->
	Pretty_utils.sfprintf  "complete_behaviors%a" pp_names lb
    | Property.IPDisjoint (_, _, lb) ->
	Pretty_utils.sfprintf  "disjoint_behaviors%a" pp_names lb
    | Property.IPDecrease (_,Kglobal,_,_) -> "decrease"
    | Property.IPDecrease (_,Kstmt _,_,_) -> "loop_variant"
    | Property.IPAxiom name -> ("axiom_"^name)
    | Property.IPAssigns (_kf, ki, _bhv, _) ->
        let name = match ki with
          | Kglobal -> "function_assigns"
          | Kstmt s ->
              match s.skind with
                | Loop _ -> "loop_assigns_" ^ string_of_int s.sid
                | _ -> "stmt_assigns_" ^ string_of_int s.sid
        in name
    | Property.IPFrom _ -> "from"
    | Property.IPBehavior _ -> ""
    | Property.IPBlob s -> State.get_name s

let prop_main p =
  match p.p_kind , p.p_prop with
  | PKProp , Property.IPAssigns (_kf, (Kstmt s), _, _) ->
      "stmt_assigns_" ^ string_of_int s.sid
  | PKProp , p -> id_prop_txt p
  | PKPropLoop , Property.IPAssigns (_kf, (Kstmt s), _, _) ->
      "loop_assigns_" ^ string_of_int s.sid
  | PKPropLoop , p -> id_prop_txt p
  | PKEstablished , p -> id_prop_txt p ^ "_established"
  | PKPreserved , p -> id_prop_txt p ^ "_preserved"
  | PKVarDecr , p -> id_prop_txt p ^ "_decrease"
  | PKVarPos , p -> id_prop_txt p ^ "_positive"
  | PKAFctOut , _ -> "normal_assigns"
  | PKAFctExit , _ -> "exit_assigns"
  | PKPre(kf,stmt,pid) , _ -> 
    Printf.sprintf "pre%d_%s_s%d" pid (Kernel_function.get_name kf) stmt.sid

let prop_txt p =
  match p.p_part with
    | None -> prop_main p
    | Some(k,_) -> Printf.sprintf "%s_part%d" (prop_main p) (succ k)

let label_of_kind = function
  | PKProp -> "Proof Obligation"
  | PKPropLoop -> "Invariant as hypothesis" (* should be assert false ??? *)
  | PKEstablished -> "Establishment"
  | PKPreserved -> "Preservation"
  | PKVarDecr -> "Decreasing"
  | PKVarPos -> "Positive"
  | PKAFctOut -> "Function assigns"
  | PKAFctExit -> "Exit assigns"
  | PKPre(kf,_,_) -> Printf.sprintf "Precondition for '%s'" (Kernel_function.get_name kf)

let label_of_prop_id p =
  match p.p_part with
    | None -> label_of_kind p.p_kind
    | Some(k,n) -> Printf.sprintf "%s (%d/%d)" (label_of_kind p.p_kind) (succ k) n

let pp_id fmt pid = Format.fprintf fmt "%s" (prop_txt pid)

let () = pretty := pp_id

let kind_order = function
  | PKProp -> 0
  | PKPre _ -> 0
  | PKEstablished -> 1
  | PKPreserved -> 2
  | PKVarPos -> 3
  | PKVarDecr -> 4
  | PKPropLoop -> 5
  | PKAFctOut -> 6
  | PKAFctExit -> 7

let compare_prop_id pid1 pid2 =
  (* This order of compatison groups together prop_pids with same properties *)
  let p1 = property_of_id pid1 in
  let p2 = property_of_id pid2 in
  let cmp = Property.compare p1 p2 in
  if cmp <> 0 then cmp
  else
    let cmp = kind_order pid2.p_kind - kind_order pid1.p_kind in
    if cmp <> 0 then cmp
    else
      Pervasives.compare pid1.p_part pid2.p_part

let set_unreachable pid =
  Wp_parameters.debug ~level:2
    "[strategy] unreachable annotation %a@." pp_id pid;
  let ip = property_of_id pid in
    Wp_parameters.feedback "[WP:unreachable] Goal %a : Valid" pp_id pid;
  WPUnreachable.set_checked ip [] True

(*----------------------------------------------------------------------------*)
(* Strategy and annotations                                                   *)
(*----------------------------------------------------------------------------*)

(*
 * 2 kinds of annotations can be found in the strategy : predicates and assigns
 * because assigns properties can only be translated into predicates
 * by the memory model.
 * - Assigns properties are composed of the assigns list from Cil,
 * and a label to know where to stop.
 * - Predicates are just the predicate type from Cil.
 *)

type t_pred = Cil_types.predicate named
type a_fun = Assigns_FctOut | Assigns_FctExit | Assigns_Stmt | Assigns_Loop
type a_kind = LoopAssigns | StmtAssigns
type t_assigns = {
  a_label : Cil_types.logic_label ;
  a_fun : a_fun ;
  a_kind : a_kind ;
  a_assigns : Cil_types.identified_term Cil_types.assigns ;
}

type t_axiom = string * Cil_types.logic_label list * Cil_types.predicate named

type t_id_pred = prop_id * t_pred
type t_id_axiom = prop_id * t_axiom
type t_id_assigns = prop_id * t_assigns

type assigns_info =
    AssignsLocations of t_id_assigns
  | AssignsAny of Cil_types.logic_label * a_kind
  | NoAssignsInfo

let body_of_id_pred = snd
let id_of_id_pred = fst

let body_of_id_assigns = snd
let id_of_id_assigns = fst

(* An annotation can be used for different purpose. *)
type annot_kind =
  | Ahyp  (* annotation is an hypothesis,
             but not a goal (see Aboth) : A => ...*)
  | Agoal (* annotation is a goal,
             but not an hypothesis (see Aboth): A /\ ...*)
  | Aboth (* annotation is both hypothesis and goal :
             to be used as A /\ A=>..*)
  | Acut  (* annotation is use as a cut : A (+ proof obligation A => ...) *)
  | Ainv  (* annotation is an invariant (not a loop invariant) :
             it will be transformed in Ahyp and Agoal by calculus. *)
  | AcallHyp (* annotation is a called function property to consider as an Hyp.
              *)
  | AcallPre (* annotation is a called function precondition : 
                it can be considered both as hyp and goal *)

let new_loops = false

(* Some elements can be used as both Hyp and Goal : because of the selection
 * mecanism, we need to add a boolean [as_goal] to tell if the element is to be
 * considered as a goal. If [false], the element can still be used as hypthesis.
 *)
type t_annots = {
  p_hyp : t_id_pred list;
  p_goal : t_id_pred list;
  p_both : (bool * t_id_pred) list;
  p_cut : (bool * t_id_pred) list;
  p_inv : (bool * t_id_pred) list;
  call_hyp : t_id_pred list; (* post and pre *)
  call_pre : t_id_pred list; (* goal only *)
  a_goal : assigns_info;
  a_hyp : assigns_info;
  a_call : assigns_info;
}

let get_goal_only annots = annots.p_goal

let get_hyp_only annots =
  let l = annots.p_both in
  let l = if new_loops then l @ annots.p_inv else l in
  let add_both_hyp acc (goal, p) = if goal then acc else p::acc in
    List.fold_left add_both_hyp annots.p_hyp l

let get_hyp_goal annots =
  let l = annots.p_both in
  let l = if new_loops then l @ annots.p_inv else l in
  let l = List.filter (fun (goal, _p) -> goal) l in
    List.map (fun (_,p) -> p) l

let get_call_hyp annots = annots.call_hyp
let get_call_pre_goal annots = annots.call_pre

let get_cut annots = annots.p_cut

let get_inv annots =
    let l = List.filter (fun (goal, _p) -> goal) annots.p_inv in
      List.map (fun (_,p) -> p) l

let get_asgn_hyp annots = annots.a_hyp

let get_asgn_goal annots = annots.a_goal

let get_call_asgn annots = annots.a_call

let pp_annots fmt acc =
  let pp_pred k b p =
    Format.fprintf fmt "%s%s: %a@."
      k (if b then "" else " (h)") !Ast_printer.d_predicate_named p
  in
  let pp_pred_list k l = List.iter (fun (_,p) -> pp_pred k true p) l in
  let pp_pred_b_list k l = List.iter (fun (b, (_,p)) -> pp_pred k b p) l in
  let pp_assign_info k a = match a with
    | NoAssignsInfo -> ()
    | AssignsAny (label,kind) -> 
      let pkind = 
        match kind with
            StmtAssigns -> ""
          | LoopAssigns -> "loop"
      in
      Format.fprintf fmt "%s(@@%a): %s assigns everything@."
        k Wp_error.pp_logic_label label pkind
    | AssignsLocations (_,a) -> Format.fprintf fmt "%s(@@%a): %a@." k
                      Wp_error.pp_logic_label a.a_label
                      Wp_error.pp_assigns a.a_assigns
  in pp_pred_list "H" acc.p_hyp;
     pp_pred_list "G" acc.p_goal;
     pp_pred_b_list "H+G" acc.p_both;
     pp_pred_b_list "C" acc.p_cut;
     pp_pred_b_list "I" acc.p_inv;
     pp_pred_list "CallHyp" acc.call_hyp;
     pp_pred_list "GoalPre" acc.call_pre;
     pp_assign_info "HA" acc.a_hyp;
     pp_assign_info "GA" acc.a_goal;
     pp_assign_info "CallA" acc.a_call;

(* This is an Hashtbl where some predicates are stored on CFG edges.
 * On each edge, we store (hypotheses, goals).
 *)
module Hannots = Cil2cfg.HE (struct type t = t_annots end)

(* This is to code what kind of properties we want to process. *)
type assigns = NoAssigns | OnlyAssigns | WithAssigns

(* This is to code which behavior the computed strategy refers to. *)
type asked_bhv =
  | FunBhv of string
  | StmtBhv of Cil2cfg.node * stmt * string

let name_of_asked_bhv = function
  | FunBhv bhv -> bhv
  | StmtBhv (_, _, bhv) -> bhv


(* This is to code what properties the user asked for in a given behavior. *)
type asked_prop =
  | AllProps
  | NamedProp of string
  | IdProp of Property.t

(* a table to keep the information about the statement default specification
 * associated with each edge in order to know in which strategy we should put a
 * default annotation on this edge. When an edge has no information in the table,
 * it means that the edge annotations belong to the [FunBhv] default behavior;
 * and when we find a statement [s], it means that they belong to the [StmtBhv s]
 * default behavior. The [int] information is only useful to build the table :
 * when an edge is included in 2 different [StmtBhv] we only keep the one that
 * has the fewer internal edges because it is necessarily included in the other.
 *)
module HdefAnnotBhv = Cil2cfg.HE (struct type t = (stmt * int) end)

type which_strategy = WSannots | WSfroms

(* Finally, a configuration is associated to a strategy computation to
 * summarize what is to be computed. *)
type strategy_info = {
  kf : Kernel_function.t;
  cfg : Cil2cfg.t;
  which_strategy : which_strategy;
  new_loops : bool;
  cur_bhv : asked_bhv;
  asked_bhvs : asked_bhv list;
  asked_prop : asked_prop;
  assigns_filter : assigns;
  def_annots_info : HdefAnnotBhv.t;
}

let new_loop_computation config = config.new_loops

type strategy_for_froms = {
  get_pre : unit -> t_annots
}
type strategy_kind =
  | SKannots (* normal mode for annotations *)
  | SKfroms of strategy_for_froms

(* an object of this type is the only access to annotations
 * from the rest of the application.
 * The idea is to be able to tune which properties to use for a computation. *)
type strategy = {
  behavior_name : string option ;
  info : strategy_info;

  has_asgn_goal : bool;
  has_prop_goal : bool;
  has_inv_prop : Cil2cfg.edge list;

  get_annots : Cil2cfg.edge -> t_annots ;

  global_axioms : t_id_axiom list ;

  strategy_kind : strategy_kind;
}

let has_asgn_goal = ref false
let has_prop_goal = ref false

let cfg_of_strategy strat = strat.info.cfg

(*----------------------------------------------------------------------------*)
(* Adding things in the stategy                                               *)
(*----------------------------------------------------------------------------*)

(* Select annotations to take as Hyp/Goal/... *)

let pp_assigns_mode fmt config =
  let str = match config.assigns_filter with
  | NoAssigns -> "without assigns"
  | OnlyAssigns -> "only with assigns"
  | WithAssigns -> "both assigns or not"
  in Format.fprintf fmt "%s" str

let pp_asked_prop fmt config = match config.asked_prop  with
  | AllProps -> Format.fprintf fmt "all properties"
  | NamedProp s ->  Format.fprintf fmt "properties %s" s
  | IdProp p -> Format.fprintf fmt "property %s" (id_prop_txt p)

let pp_strategy_info fmt config =
  Format.fprintf fmt "'%a', " Kernel_function.pretty_name config.kf;
  let _ = match config.cur_bhv with
    | FunBhv bhv ->
        Format.fprintf fmt "behavior '%s'" bhv
    | StmtBhv (_, s, bhv) ->
        Format.fprintf fmt "behavior '%s' of statement %d" bhv s.sid
  in Format.fprintf fmt ", %a, %a"
       pp_asked_prop config pp_assigns_mode config

let cur_fct_default_bhv config = match config.cur_bhv with
  | FunBhv bhv -> bhv =  Cil.default_behavior_name
  | _ -> false

let filter_status id =
  match fst (Properties_status.strongest (property_of_id id)) with 
    | Checked { valid=True } -> Wp_parameters.StatusTrue.get ()
    | Checked { valid=Maybe } -> Wp_parameters.StatusMaybe.get ()
    | Checked { valid=False } -> Wp_parameters.StatusFalse.get ()
    | Unknown -> true

let goal_to_select config pid =
  let asked, take_it =
    match config.assigns_filter, property_of_id pid with
      | NoAssigns, Property.IPAssigns _ ->
          "no assigns", false
      | (OnlyAssigns | WithAssigns), Property.IPAssigns _ ->
          "", true
      | OnlyAssigns, _ -> "only assigns", false
      | (NoAssigns | WithAssigns), _ -> "", true
  in
  let asked, (take_it, msg) =
    if take_it then
      begin match config.asked_prop with
    | AllProps ->
        "all",
        if (filter_status pid) 
        then true, " (selected)"
	else false, " (skipped w.r.t status)"
    | IdProp idp ->
        (* Notice that if the user has explicitly selected a property,
        * we consider it as a goal even if it has been proved already *)
        (id_prop_txt idp),
      (Property.equal (property_of_id pid) idp, "")
    | NamedProp str ->
        str,
        if List.mem str (user_prop_names pid)
        then true, " (asked named prop)" else false, ""
      end
    else asked, (take_it, "")
  in
    Wp_parameters.debug ~level:3 "[goal_to_select] %s vs %a -> %s%s@."
        asked pp_id pid (if take_it then "select" else "ignore") msg;
    take_it

(*--- Low level functions ---*)
(* to build identifications, normalize the labels and add to accumulators. *)

(* Functions that finally add things *)

let empty_acc =
  { p_hyp = []; p_goal = []; p_both = []; p_cut = []; p_inv = [];
    call_hyp = []; call_pre = []; a_call = NoAssignsInfo;
    a_goal = NoAssignsInfo; a_hyp = NoAssignsInfo; }

(* TODO: it should be possible to do without this, but needs a big refactoring*)
let merge_acc acc1 acc2 = 
  let merge_assigns a1 a2 kind =
    match a1,a2 with
      | NoAssignsInfo, a | a, NoAssignsInfo -> a
      | (AssignsLocations _ | AssignsAny _), 
        (AssignsLocations _ | AssignsAny _) ->
        Wp_parameters.fatal "Several assigns %s ?" kind
in
{
  p_hyp = acc1.p_hyp @ acc2.p_hyp;
  p_goal = acc1.p_goal @ acc2.p_goal;
  p_both = acc1.p_both @ acc2.p_both;
  p_cut = acc1.p_cut @ acc2.p_cut;
  p_inv = acc1.p_inv @ acc2.p_inv;
  call_hyp = acc1.call_hyp @ acc2.call_hyp;
  call_pre = acc1.call_pre @ acc2.call_pre;
  a_goal = merge_assigns acc1.a_goal acc2.a_goal "goals";
  a_hyp = merge_assigns acc1.a_hyp acc2.a_hyp "hyp";
  a_call = merge_assigns acc1.a_call acc2.a_call "hyp"
}

let add_prop config acc kind (labels, id, p) =
  let get_p () =
    let txt = prop_txt id in
      try
        let p = NormAtLabels.preproc_annot labels p in
        let _ =
          Wp_parameters.debug ~level:3 "take (@[%s:@ %a@])@."
            txt !Ast_printer.d_predicate_named p
        in Some (id, p)
      with e -> NormAtLabels.catch_label_error e txt "annotation"; None
  in
  let add_hyp l = match get_p () with None -> l | Some p -> p::l in
  let add_goal l =
    if goal_to_select config id
    then match get_p () with None -> l
      | Some p -> ( has_prop_goal := true; p::l )
        else l
  in
  let add_both l =
    match get_p () with None -> l
      | Some p ->
          let goal = goal_to_select config id in
            if goal then has_prop_goal := true;
            (goal, p)::l
  in
    match kind with
      | Ahyp -> { acc with p_hyp = add_hyp acc.p_hyp }
      | Agoal -> { acc with p_goal = add_goal acc.p_goal }
      | Aboth -> { acc with p_both = add_both acc.p_both }
      | Acut -> { acc with p_cut = add_both acc.p_cut }
      | Ainv -> { acc with p_inv = add_both acc.p_inv }
      | AcallHyp -> { acc with call_hyp = add_hyp acc.call_hyp }
      | AcallPre -> { acc with call_pre = add_goal acc.call_pre }

(* Add properties *)

let add_glob_axiom name acc labels a =
  try
    let a = NormAtLabels.preproc_annot NormAtLabels.labels_axiom a in
    let labels =
      List.map (NormAtLabels.preproc_label NormAtLabels.labels_axiom) labels in
    let id = mk_axiom_id name in
      Wp_parameters.debug ~level:3 "take (@[%s:@ %a@])@."
        (prop_txt id)
        !Ast_printer.d_predicate_named a;
      (id, (name,labels,a))::acc
  with e -> NormAtLabels.catch_label_error e ("axiom "^name) "axiom"; acc

let prepare_pre_pred config ki bhv pre p =
  let prop_id = mk_pre_id config.kf ki bhv pre in
  let p, labels = match ki with
    | Kglobal -> Logic_const.pat (p, Logic_const.pre_label), NormAtLabels.labels_fct_pre
    | Kstmt s -> p, NormAtLabels.labels_stmt_pre s
  in (labels, prop_id, p)

let prepare_pre config ki bhv pre =
  let p = Logic_const.pred_of_id_pred pre in
    prepare_pre_pred config ki bhv pre p

let prepare_fct_post config bhv tkind _assumes post =
  let p = Logic_const.pred_of_id_pred post in
  (*let assumes = Logic_const.pold assumes in
  let p = Logic_const.pimplies (assumes, p) in *)
  let prop_id = mk_fct_post_id config.kf bhv (tkind, post) in
  let labels = NormAtLabels.labels_fct_post in
    (labels, prop_id, p)

let prepare_stmt_spec_post config bhv s l_post assumes tkind post =
  let p = Logic_const.pred_of_id_pred post in
  let assumes = Logic_const.pold assumes in
               (* can use old because label normalisation will be called *)
  let p = Logic_const.pimplies (assumes, p) in
  let prop_id = mk_stmt_post_id config.kf s bhv (tkind, post) in
  let labels = NormAtLabels.labels_stmt_post s l_post in
    (labels, prop_id, p)

let prepare_var_decr config s ca inv =
  let id = mk_var_decr_id config.kf s ca in
  let labels = NormAtLabels.labels_loop_inv s in
    (labels, id, inv)

let prepare_var_pos config s ca inv =
  let id = mk_var_pos_id config.kf s ca in
  let labels = NormAtLabels.labels_loop_inv s in
    (labels, id, inv)

let prepare_inv_establish config s ca p =
  let id = mk_establish_id config.kf s ca in
  let labels = NormAtLabels.labels_loop_inv s in
    (labels, id,p)

let prepare_inv_preserve config s ca p =
  let id = mk_preserve_id config.kf s ca in
  let labels = NormAtLabels.labels_loop_inv s in
    (labels, id,p)

let prepare_inv_fixpoint config s ca p =
  let id = mk_inv_hyp_id config.kf s ca in
  let labels = NormAtLabels.labels_loop_inv s in
    (labels, id, p)

(* This is for complete/disjoint behaviors *)
let prepare_behaviors_spec ki id a =
  let labels = match ki with
    | Kglobal -> NormAtLabels.labels_fct_pre
    | Kstmt s -> NormAtLabels.labels_stmt_pre s
  in
    (labels, id, a)

let prepare_annot before_after config s l_post ca pred =
  let id = mk_code_annot_id config.kf s ca in
  let labels =
    if before_after then NormAtLabels.labels_assert_before s
    else NormAtLabels.labels_assert_after s l_post
  in (labels, id, pred)

(* Add Assigns *)

let add_assigns config kind ki bhv_opt src_asgns acc t_asgns =
  let id =
    match t_asgns.a_fun with
      | Assigns_Stmt -> mk_stmt_assigns_id (config.kf, ki, bhv_opt, src_asgns)
      | Assigns_Loop -> mk_loop_assigns_id (config.kf, ki, bhv_opt, src_asgns)
      | Assigns_FctOut -> mk_fout_assigns_id (config.kf, ki, bhv_opt, src_asgns)
      | Assigns_FctExit -> mk_exit_assigns_id (config.kf, ki, bhv_opt, src_asgns)
  in
  let take_assigns () =
    Wp_parameters.debug ~level:3 
      "take %a %a" pp_id id Wp_error.pp_assigns (Writes src_asgns);
    AssignsLocations (id, t_asgns)
  in
  match kind with
    | Ahyp -> {acc with a_hyp = take_assigns ()}
    | AcallHyp -> {acc with a_call = take_assigns ()}
    | Agoal ->
        if (goal_to_select config id) then
          ( has_asgn_goal:= true; {acc with a_goal = take_assigns ()} )
        else acc
    | _ -> Wp_parameters.fatal "Assigns prop can only be Hyp or Goal"

(* [VP 2011-01-28] following old behavior, not sure it is correct:
   why should we give to add_assigns the assigns with unnormalized labels?
*)
let add_loop_assigns config kind s (ca,assigns) acc =
  let labels = NormAtLabels.labels_loop_assigns s in
  let a = NormAtLabels.preproc_assigns labels assigns in
    add_assigns config kind (Kstmt s) (Property.Id_code_annot ca) assigns acc
      {
        a_label = Clabels.mk_logic_label s ;
        a_fun = Assigns_Loop ;
        a_kind = LoopAssigns ;
        a_assigns = Writes a
      }

let ki_assigns_labels ki l_post =
  let l_begin, norm_labels = match ki with
    | Kglobal ->
        Logic_const.pre_label, NormAtLabels.labels_fct_assigns
    | Kstmt s ->
        Clabels.mk_logic_label s, NormAtLabels.labels_stmt_assigns s l_post
  in l_begin, norm_labels


let add_bhv_assigns config kind ki acc a_fun b l_post =
  let l, labels = ki_assigns_labels ki l_post in
  match b.b_assigns with
    | WritesAny -> 
      (match kind with
        | AcallHyp -> { acc with a_call = AssignsAny (l,StmtAssigns) }
        | Ahyp -> { acc with a_hyp = AssignsAny (l,StmtAssigns) }
        | _ -> acc)
    | Writes assigns ->
        add_assigns config kind ki (Property.Id_behavior b) assigns acc {
	  a_label = l ; a_fun = a_fun ; a_kind = StmtAssigns ;
	  a_assigns = Writes (NormAtLabels.preproc_assigns labels assigns) ;
	}

let assigns_upper_bound bhvs =
  let upper a b =
    match a,b.b_assigns with
        | None, Writes a when Cil.is_default_behavior b -> 
          Some (b,a) (* default behavior always applies. *)
        | None, _ -> None
        | Some (b,_), _ when Cil.is_default_behavior b -> 
          a (* default behavior prevails over other behaviors. *)
        | Some _, WritesAny -> 
          None (* No default behavior and one behavior assigns everything. *)
        | Some(b,a1), Writes a2 -> Some (b,a1 @ a2)
          (* take the whole list of assigns. *)
  in
  match bhvs with
    | [] -> None
    | bhv::bhvs ->
      (* [VP 2011-02-04] Note that if there is no default and each
         behavior has a proper assigns clause we put dependencies only
         to the assigns of a more or less randomly selected behavior,
         but the datatypes above can't handle anything better.  *)
      let acc = 
        match bhv.b_assigns with
            WritesAny -> None
          | Writes a -> Some(bhv,a)
      in
      List.fold_left upper acc bhvs

(* [VP 2011-02-04] These two functions below mix all the assigns of 
   a function regardless of the behavior. At least now that we take 
   WritesAny as soon as at least one behavior has no assigns clause,
   this is correct, but still imprecise. Needs refactoring of t_annots to
   go further, though.
 *)
let add_spec_assigns_as_hyp config ki l_post acc spec =
  let l,labels = ki_assigns_labels ki l_post in
  match assigns_upper_bound spec.spec_behavior with
    | None -> { acc with a_hyp = AssignsAny (l,StmtAssigns)}
    | Some(bhv, assigns) ->
      let assigns = NormAtLabels.preproc_assigns labels assigns in
      add_assigns config Ahyp ki (Property.Id_behavior bhv) assigns acc 
        {
          a_label = l ;
          a_fun = Assigns_Stmt ;
          a_kind = StmtAssigns ;
          a_assigns = Writes assigns ;
        }

let add_call_assigns config s acc spec_opt l_post =
  let l,labels = ki_assigns_labels (Kstmt s) l_post in
  match spec_opt with
    | None -> { acc with a_call = AssignsAny (l,StmtAssigns) }
    | Some spec -> 
      match assigns_upper_bound spec.spec_behavior with
        | None -> { acc with a_call = AssignsAny (l,StmtAssigns)}
        | Some(bhv, assigns) ->
          let assigns = NormAtLabels.preproc_assigns labels assigns in
          add_assigns 
            config AcallHyp (Kstmt s) 
            (Property.Id_behavior bhv) assigns acc 
            {
              a_label = l          ;
              a_fun = Assigns_Stmt ;
              a_kind = StmtAssigns ;
              a_assigns = Writes assigns  ;
            }

(* ------------------------------------------------------------------------ *)
(* --- Get annotations according to the behavior                        --- *)
(* ------------------------------------------------------------------------ *)

let get_named_bhv name bhv_list =
  try Some (List.find (fun b -> b.b_name = name) bhv_list)
  with Not_found -> None

(* Select in [bhv_list] the behavior that has to be processed
 * according to [config] and [ki] current statement. *)
let get_behav config ki bh_list = match config.cur_bhv, ki with
  | FunBhv b, Kglobal -> get_named_bhv b bh_list
  | StmtBhv (_, s1, b), Kstmt s2 when s1.sid = s2.sid ->
      get_named_bhv b bh_list
  | _ -> None

(* @return the names of the behavior of spec,
 * and add the default behavior name when [with_default]. *)
let get_behavior_names ~with_default spec =
  let rec get_named_bhv lb = match lb with [] -> []
    | b::tlb ->
        if Cil.is_default_behavior b
        then (* do it later*) get_named_bhv tlb
        else (b.b_name)::(get_named_bhv tlb)
  in let named_bhv = get_named_bhv spec.spec_behavior in
    if with_default then Cil.default_behavior_name::named_bhv else named_bhv

(* Tells weather the property belonging to the behaviors in [bhv_name_list]
 * has to be considered according to [config]. *)
type test_behav_res =
  | TBRno   (* [cur_bhv] is not concerned *)
  | TBRhyp  (* the property belongs to [default_behavior],
               but not to [cur_bhv] : it doesn't have to be a Goal
               but can be considered as an hypothesis. *)
  | TBRpart (* the property has to be taken as a Goal, but even if it is
               proved for every [asked_bhvs], it will still be a partial proof.
               TODO: use this to generate PKPartial ! *)
  | TBRok   (* Select as a Goal *)

(* (see [test_behav_res] above).
 * If the annotation doesn't have "for" names, it is a bit complicated because
 * we have to know if the statement [s] is inside a stmt behavior or not. *)
let is_annot_for_config config edges_before s_annot bhv_name_list =
  Wp_parameters.debug ~level:3 "[is_annot_for_config] at sid:%d for %a ? @."
    s_annot.sid (Wp_error.pp_string_list ~sep:" " ~empty:"<default>")
    bhv_name_list;
  let hyp_but_not_at_post n = (* don't take assert at post pgpt (see #564) *)
    let s_post =
      try
        let next =  Cil2cfg.node_after config.cfg n in
        match Cil2cfg.pred_e config.cfg next with
            | e::_ -> Cil2cfg.get_edge_next_stmt config.cfg e
            | _ -> None
      with Not_found -> None
    in match s_post with
      | Some s_post when s_post.sid = s_annot.sid ->  TBRno
      | _ -> TBRhyp
  in
  let res = match bhv_name_list with
    | [] -> (* no spec 'for' in the property *)
        begin
          let e = match edges_before with
            | e::_ -> e
            | _ -> Wp_parameters.fatal "annot with no edge ?"
          in
          match config.cur_bhv with
            | FunBhv b when b = Cil.default_behavior_name ->
                begin
                  try
                    let _ = HdefAnnotBhv.find config.def_annots_info e in
                      TBRhyp
                  with Not_found -> TBRok
                end
            | StmtBhv (n, sb, bname) when bname = Cil.default_behavior_name ->
                begin
                  try
                    let s,_ = HdefAnnotBhv.find config.def_annots_info e in
                      if s.sid = sb.sid then TBRok
                      else raise Not_found
                  with Not_found -> hyp_but_not_at_post n
                end
            | FunBhv _ -> TBRhyp
            | StmtBhv (n,_,_) -> hyp_but_not_at_post n
        end
    | bhvs -> (* TODOopt : there is surely a better way to do this : *)
      let asked_bhv = name_of_asked_bhv config.cur_bhv in
      let goal = List.exists (fun bl -> bl = asked_bhv) bhvs in
          if goal then
            let full = (* TODO *) true
              (* List.for_all (fun bl -> is_in bl config.asked_bhvs) bhvs *)
            in (if full then TBRok else TBRpart)
          else TBRno
  in let s_res = match res with
    | TBRok -> "ok"
    | TBRhyp -> "hyp"
    | TBRno -> "no"
    | TBRpart -> "part"
  in Wp_parameters.debug ~level:3 "[is_annot_for_config] -> %s@." s_res;
     res

let build_bhv_assumes b =
  let assumes = List.map Logic_const.pred_of_id_pred b.b_assumes in
  let assumes = Logic_const.pands assumes in
    assumes

let get_bhv_assumes spec l =
  let rec get_assumes bhv_names = match bhv_names with [] -> []
    | bhv::tl ->
        let l = match get_named_bhv bhv spec.spec_behavior with
          | None -> Wp_parameters.warning "no %s behavior !?!?" bhv;
                    get_assumes tl
          | Some b ->
                (build_bhv_assumes b)::(get_assumes tl)
        in l
  in get_assumes l

let add_fct_pre config acc spec =
  let ki = Kglobal in
  let kind = (*TODO[LC] restablish when pre-cond is OK.
	       if is_main_entry_point config.kf then Aboth else *) Ahyp in
  let add b acc p = 
    add_prop config acc kind (prepare_pre config ki b p) 
  in
  let acc = match Cil.find_default_behavior spec with None -> acc
    | Some b -> List.fold_left (add b) acc b.b_requires
  in
  let acc = match get_behav config ki spec.spec_behavior with None -> acc
    | Some b ->
	let acc =
	  if not (Cil.is_default_behavior b) then
	    List.fold_left (add b) acc b.b_requires
	  else acc (* since default requires are already added *)
	in List.fold_left (add b) acc b.b_assumes
  in acc

let get_fct_post config tkind spec =
  match get_behav config Kglobal spec.spec_behavior with
    | None -> empty_acc
    | Some b ->
        let assumes = build_bhv_assumes b in
        let add acc (tk, p) = match tkind, tk with
          | Normal, Normal ->
              add_prop config acc Agoal
                (prepare_fct_post config b Normal assumes p)
          | Exits, Exits ->
              add_prop config acc Agoal
                (prepare_fct_post config b Exits assumes p)
          | Normal, (Breaks|Continues|Returns) ->
              (Wp_parameters.warning
                 "Local abrupt terminaison property in function spec ignored :@, %a"
                 Cil.defaultCilPrinter#pPost_cond (tk,p);
               acc)
          | _ -> acc
        in
          List.fold_left add empty_acc b.b_post_cond

let get_variant spec =
  match spec.spec_variant with None -> ()
    | Some v ->
        Wp_parameters.warning "Ignored 'decrease' specification:@, %a@."
                   Cil.d_decreases v

let get_terminates spec =
  match spec.spec_terminates with None -> ()
    | Some p ->  Wp_parameters.warning
                   "ignored 'terminates' specification:@, %a@."
                   !Ast_printer.d_predicate_named
                   (Logic_const.pred_of_id_pred p)


let add_disjoint_behaviors_props config ki spec acc =
  match spec.spec_disjoint_behaviors with [] -> acc
    | l ->
        let mk_disj_bhv b1 b2 = (* ~ (b1 /\ b2) *)
          let p = Logic_const.pands [b1; b2] in
            Logic_const.pnot p
        in
        let add_disj acc bhv_names =
          let id = (config.kf, ki, bhv_names) in
          let id = mk_disj_bhv_id id in
          let bhv_names = match bhv_names with
            | [] -> (* no names ==> all behaviors *)
                get_behavior_names ~with_default:false spec
            | _ -> bhv_names
          in
          let bhv_assumes = get_bhv_assumes spec bhv_names in
          let do_one_with_list prop b lb =
            let lp = List.map (mk_disj_bhv b) lb in
              Logic_const.pands (prop::lp)
          in
          let rec do_list prop l = match l with [] -> prop
            | b::tl ->
                let prop = do_one_with_list prop b tl in
                  do_list prop tl
          in let prop = do_list Logic_const.ptrue bhv_assumes in
            add_prop config acc Agoal (prepare_behaviors_spec ki id prop)
        in List.fold_left add_disj acc l

let add_complet_behaviors_props config ki spec acc =
  match spec.spec_complete_behaviors with [] -> acc
    | l ->
        let mk_prop acc bhv_names =
          let id = (config.kf, ki, bhv_names) in
          let id = mk_compl_bhv_id id in
          let bhv_names = match bhv_names with
            | [] -> (* no names ==> all behaviors *)
                get_behavior_names ~with_default:false spec
            | _ -> bhv_names
          in
          let bhv_assumes = get_bhv_assumes spec bhv_names in
          let p = Logic_const.pors bhv_assumes in
            add_prop config acc Agoal (prepare_behaviors_spec ki id p)
        in
          List.fold_left mk_prop acc l

let add_behaviors_props config ki spec acc =
  let add = match config.cur_bhv, ki with
    | FunBhv b, Kglobal when b = Cil.default_behavior_name -> true
    | StmtBhv (_, cur_s,  b), Kstmt s
        when (s.sid = cur_s.sid && b = Cil.default_behavior_name) -> true
    | _ -> false
  in
    if add then
      let acc = add_complet_behaviors_props config ki spec acc in
      let acc = add_disjoint_behaviors_props config ki spec acc in
        acc
    else acc

(* Process the stmt spec precondition as an hypothesis for external properties.
 * Another processing must be done to prove this precondition. *)
let add_stmt_spec_precond_as_hyp config s spec acc =
  let add_bhv_pre acc b =
    let assumes = build_bhv_assumes b in
    let add acc pre =
      let p = Logic_const.pred_of_id_pred pre in
      let p = Logic_const.pimplies (assumes, p) in
        add_prop config acc Ahyp (prepare_pre_pred config (Kstmt s) b pre p)
    in List.fold_left add acc b.b_requires
  in List.fold_left add_bhv_pre acc spec.spec_behavior

let add_stmt_spec_postcond config s l_post b kind acc =
  let assumes = build_bhv_assumes b in
  let add_ensure (p_acc, e_acc) e = match e with
    | Normal, e ->
        let a = prepare_stmt_spec_post config b s l_post assumes Normal e in
        let p_acc = add_prop config p_acc kind a in
          (p_acc, e_acc)
    | Exits, e ->
        let a = prepare_stmt_spec_post config b s l_post assumes Exits e in
        let e_acc = add_prop config e_acc kind a in
          (p_acc, e_acc)
    | (Breaks|Continues|Returns),_ -> (* TODO *)
        begin
          if kind = Agoal then
            Wp_parameters.warning
              "Abrupt statement termination property ignored:@, %a"
              (Cil.defaultCilPrinter)#pPost_cond e;
          acc
        end
  in
  List.fold_left add_ensure acc b.b_post_cond

let add_stmt_spec_before config s spec acc =
  let ki = (Kstmt s) in
  match config.cur_bhv with
  | StmtBhv (_n, cur_s, _cur_bhv) when s.sid = cur_s.sid ->
      begin match get_behav config ki spec.spec_behavior with
        | None -> (* in some cases, it seems that we can have several spec
                     for the same statement -> not an error *)
            acc
        | Some b ->
              let add kind acc p =
                add_prop config acc kind (prepare_pre config ki b p)
              in
              let acc =  List.fold_left (add Aboth) acc b.b_requires in
              let acc =  List.fold_left (add Ahyp) acc b.b_assumes in
              let acc = add_behaviors_props config ki spec acc in
                acc
      end
  | _ -> (* in all other cases, use the specification as hypothesis *)
        add_stmt_spec_precond_as_hyp config s spec acc

let add_stmt_spec_after config s l_post spec acc =
  let ki = (Kstmt s) in
  match config.cur_bhv with
  | StmtBhv (_n, cur_s, _cur_bhv) when s.sid = cur_s.sid ->
      begin match get_behav config ki spec.spec_behavior with
        | None -> (* in some cases, it seems that we can have several spec
                     for the same statement -> not an error *)
            acc
        | Some b ->
            let _ = get_variant spec in (* TODO *)
            let _ = get_terminates spec in (* TODO *)
            let p_acc, e_acc =
              add_stmt_spec_postcond config s l_post b Agoal acc in
            let p_acc = (* TODO: check that assigns are for post only *)
              add_bhv_assigns config Agoal ki p_acc Assigns_Stmt b l_post
            in (p_acc, e_acc)
      end
  | _ -> (* in all other cases, use the specification as hypothesis *)
      let add_bhv_post acc b =
        let p_acc, e_acc = add_stmt_spec_postcond config s l_post b Ahyp acc in
        let p_acc = add_spec_assigns_as_hyp config ki l_post p_acc spec in
        (p_acc, e_acc)
      in
      List.fold_left add_bhv_post acc spec.spec_behavior

(*----------------------------------------------------------------------------*)
(* Call annotations                                                           *)
(*----------------------------------------------------------------------------*)

let get_called_pre which kf spec =
  Wp_parameters.debug ~level:3 "[get_called_pre] for %a@."
    Kernel_function.pretty_name  kf;
  match which with
    | WSfroms -> []
    | WSannots ->
        let labels = NormAtLabels.labels_fct_pre in
        let add_behav acc b = (* pre for behavior is [assumes => requires] *)
          let assumes = Logic_const.pold (build_bhv_assumes b) in
          let add_pre acc pre =
            let id = mk_pre_id kf Kglobal b pre in
            let p = Logic_const.pred_of_id_pred pre in
            let p = Logic_const.pimplies (assumes, p) in
              (labels, id, p)::acc
          in List.fold_left add_pre acc b.b_requires
        in
        let acc = List.fold_left add_behav [] spec.spec_behavior in
          if acc = [] then
            Wp_parameters.debug ~level:3 "no called precond for %a@."
              Kernel_function.pretty_name  kf;
          acc

let get_called_post which kf termination_kind =
  let spec = Kernel_function.get_spec kf in
  Wp_parameters.debug ~level:3 "[get_called_post] '%s' for %a@."
    (string_of_termination_kind termination_kind)
    Kernel_function.pretty_name  kf;
  let labels = NormAtLabels.labels_fct_post in
  let acc = match which with 
    | WSfroms ->
        let posts = 
          WpFroms.post_of_assigns_in_spec kf Kglobal spec termination_kind in
        let mk_prop (bhv, post) =
          let p = Logic_const.pred_of_id_pred post in
          let id = mk_fct_post_id kf bhv (termination_kind, post) in
            (labels, id, p)
        in List.map mk_prop posts
    | WSannots ->
        let mk_id b p = mk_fct_post_id kf b (termination_kind, p) in
        let add_behav acc b = 
          (* post for behavior is [\old(assumes) => ensures] *)
          let assumes = Logic_const.pold (build_bhv_assumes b) in
          let add_post acc p = match p with
            | kind, p when kind = termination_kind ->
                let id = mk_id b p in
                let p = Logic_const.pred_of_id_pred p in
                let p = Logic_const.pimplies (assumes, p) in
                  (labels, id, p)::acc
            | _ -> acc
          in 
          List.fold_left add_post acc b.b_post_cond
        in
        List.fold_left add_behav [] spec.spec_behavior 
  in
    if acc = [] then
      Wp_parameters.debug ~level:3 "no called %s postcondition for %a@."
        (string_of_termination_kind termination_kind)
        Kernel_function.pretty_name  kf;
    acc

let get_call_annots config s fct l_post =
  match get_called_kf fct with
    | Some kf ->
        let spec = Kernel_function.get_spec kf in
        let add kind acc p = add_prop config acc kind p in
        let add_goal_pre acc (labelmap,pid,predicate) = 
	  setup_precondition_proxy kf pid.p_prop ;
	  let prop = {
	    p_prop = PreCondAt.property pid.p_prop s ;
	    p_kind = PKPre(kf,s,pre_cond_id pid.p_prop) ;
	    p_part = None ;
	  } in
	  add_prop config acc AcallPre (labelmap,prop,predicate) 
	in

        let pre = get_called_pre config.which_strategy kf spec in
        let before_annots = List.fold_left (add AcallHyp) empty_acc pre in
	let before_annots =
	  if rte_precond_status config.kf then before_annots 
	  else List.fold_left add_goal_pre before_annots pre in
        let normal_post = get_called_post config.which_strategy kf Normal in
        let post_annots = List.fold_left (add AcallHyp) empty_acc normal_post in
        let post_annots = 
          add_call_assigns config s post_annots (Some spec) l_post 
        in
        let exits_post = get_called_post config.which_strategy kf Exits in
        let exits_annots = List.fold_left (add AcallHyp) empty_acc exits_post in
        before_annots, (post_annots, exits_annots)

    | None ->
        Wp_parameters.warning 
          "call through function pointer not implemented yet: \
                                   ignore called function properties.";
        let assigns_annots = add_call_assigns config s empty_acc None l_post in
        empty_acc, (assigns_annots, assigns_annots)

(*----------------------------------------------------------------------------*)

let get_stmt_annots config ~loop edges_before s l_post =
  let add ((b_acc, (a_acc, e_acc)) as acc) b_list ca kind p before =
    let add_to_acc k =
      let a = prepare_annot before config s l_post ca p in
      if before then
        let b_acc = add_prop config b_acc k a in
          (b_acc, (a_acc, e_acc))
      else
        let a_acc = add_prop config a_acc k a in
          (b_acc, (a_acc, e_acc))
    in
    match is_annot_for_config config edges_before s b_list with
    | TBRok | TBRpart -> add_to_acc kind
    | TBRhyp -> add_to_acc Ahyp
    | TBRno -> acc
  in
  let add_code_annot ca_before ca ((b_acc, (a_acc, e_acc)) as acc) =
    match ca.annot_content with
      | AInvariant (_b_list, loop_inv, _inv) ->
          if loop_inv then (* see get_loop_annots *) acc
          else begin
            (* TODO : add acc b_list ca Ainv inv ca_before *)
            Wp_parameters.warning 
                 "ignored 'invariant' (not implemented yet) : %a"
                 !Ast_printer.d_code_annotation ca;
            acc
          end
      | AAssert (b_list,p) ->
          add acc b_list ca Aboth p ca_before
      | AAssigns (_b_list, _assigns) -> (* loop assigns *)
          if loop then (* see get_loop_annots *) acc
          else
            (Wp_parameters.warning "Loop assigns but not a loop. Ignored: %a"
               !Ast_printer.d_code_annotation ca;
             acc)
      | AVariant (_v, _rel) ->
          if loop then (* see get_loop_annots *) acc
          else (Wp_parameters.warning "Variant outside a loop. Ignored: %a"
                  !Ast_printer.d_code_annotation ca;
                acc)
      | APragma _ ->
          Wp_parameters.warning "Ignored annotation: %a"
            !Ast_printer.d_code_annotation ca;
          acc
      | AStmtSpec spec ->
          if ca_before then
            let b_acc = add_stmt_spec_before config s spec b_acc in
            let (a_acc, e_acc) =
              add_stmt_spec_after config s l_post spec (a_acc, e_acc)
            in  (b_acc, (a_acc, e_acc))
          else (Wp_parameters.warning ~once:true
                  "Ignoring specification rooted after statement %d" s.sid;
                acc)
  in
  let do_annot a acc = match a with
    | Before (User ca | AI (_, ca)) -> add_code_annot true  ca acc
    | After  (User ca | AI (_, ca)) -> add_code_annot false ca acc
  in
  let before_acc, after_acc, exits_acc = empty_acc, empty_acc, empty_acc in
  let acc = before_acc, (after_acc, exits_acc) in
    Annotations.single_fold_stmt do_annot s acc

let mk_variant_properties s v =
  let loc = v.term_loc in
  let lhead = Clabels.loop_head_label s in
  let vhead = Logic_const.tat ~loc (v, lhead) in
  let zero = Cil.lzero ~loc () in
  let vpos = Logic_const.prel ~loc (Rle, zero, vhead) in
  let vdecr = Logic_const.prel ~loc (Rlt, v, vhead) in
  vpos, vdecr

(* Returns the annotations for the three edges of the loop node:
 * - loop_entry : goals for the edge entering in the loop
 * - loop_back  : goals for the edge looping to the entry point
 * - loop_core  : fix-point hypothesis for the edge starting the loop core
 *)
let get_loop_annots config edges_before s =
  let add ca (assigns, loop_entry, loop_back , loop_core as acc) =
    match ca.annot_content with
      | AInvariant (b_list, true, inv) ->
          (* we have to prove that inv is true for each edge that goes
           * in the loop, so we can assume that inv is true for each edge
           * starting from this point. *)
          begin match is_annot_for_config config edges_before s b_list with
            | TBRok
            | TBRpart (* TODO: PKPartial *)
              ->
                let loop_entry =
                  add_prop config loop_entry Agoal
                    (prepare_inv_establish config s ca inv ) in
                let loop_back =
                  add_prop config loop_back Agoal
                    (prepare_inv_preserve config s ca inv ) in
                let loop_core =
                  add_prop config loop_core Ahyp
                    (prepare_inv_fixpoint config s ca inv )
                in assigns, loop_entry , loop_back , loop_core
            | TBRhyp -> (* TODO : add more inv hyp ? *)
                let loop_core =
                  add_prop config loop_core Ahyp
                    (prepare_inv_fixpoint config s ca inv)
                in assigns, loop_entry , loop_back , loop_core
            | TBRno -> acc
          end
      | AVariant (v, None) ->
          let kind = if cur_fct_default_bhv config then Agoal else Ahyp in
          let vpos, vdecr = mk_variant_properties s v in
          let vdecr = prepare_var_decr config s ca vdecr in
          let loop_back = add_prop config loop_back kind vdecr in
          let vpos = prepare_var_pos config s ca vpos in
          let loop_back = add_prop config loop_back kind vpos in
            assigns, loop_entry , loop_back , loop_core
      | AVariant (_v, _rel) ->
            Wp_parameters.warning "Ignoring loop variant with measure : %a"
              !Ast_printer.d_code_annotation ca;
            acc
      | AAssigns (_,WritesAny) -> assert false
      | AAssigns (b_list, Writes a) -> (* loop assigns *)
          let h_assigns, g_assigns = assigns in
          let check_assigns old cur =
            match old with
                None -> Some cur
              | Some _ -> 
                Wp_parameters.fatal
                  "At most one loop assigns can be associated to a behavior"
          in
          let assigns = 
            match is_annot_for_config config edges_before s b_list with
            | TBRok | TBRpart -> 
              check_assigns h_assigns (ca,a), check_assigns g_assigns (ca,a)
            | TBRhyp -> 
              check_assigns h_assigns (ca,a), g_assigns
            | TBRno -> assigns
          in (assigns, loop_entry , loop_back , loop_core)
      | _ -> acc (* see get_stmt_annots *)
  in
  let do_annot a acc = match a with
    | Before (User ca | AI (_, ca)) -> add ca acc
    | After _ -> acc (* Warning already emitted in get_stmt_annots *)
  in
  let acc = ((None,None), empty_acc, empty_acc, empty_acc) in
  let (h_assigns, g_assigns), loop_entry , loop_back , loop_core =
    Annotations.single_fold_stmt do_annot s acc
  in
  let loop_back = 
    match g_assigns with 
        None -> loop_back
      | Some a -> add_loop_assigns config Agoal s a loop_back 
  in
  let loop_core = 
    match h_assigns with 
        None -> 
          let lab = Clabels.mk_logic_label s in
          { loop_core with a_hyp = AssignsAny (lab,LoopAssigns) }
      | Some a -> add_loop_assigns config Ahyp s a loop_core 
  in
  loop_entry , loop_back , loop_core

let get_fct_assigns config a_fun spec =
  let ki = Kglobal in
  match get_behav config ki spec.spec_behavior with
    | None -> empty_acc
    | Some b -> add_bhv_assigns config Agoal ki empty_acc a_fun b None

(*----------------------------------------------------------------------------*)
(* Build graph annotation for the strategy                                    *)
(*----------------------------------------------------------------------------*)

let add_on_edges annots new_acc edges =
  let add_on_edge e =
    let acc =
      try let acc = Hannots.find annots e in merge_acc new_acc acc
      with Not_found -> new_acc
    in Hannots.replace annots e acc
  in List.iter add_on_edge edges

(* Builds tables that give hypotheses and goals relative to [b] behavior
 * for edges of the cfg to consider during wp computation.
 * [b = None] means that we only consider internal properties to select for the
 * default behavior. This is useful when the function doesn't have any
 * specification.
 * @param asked_prop = Some id -> select only this goal (use all hyps).
 *)
let get_behavior_annots cfg config =
  Wp_parameters.feedback ~level:3
    "build strategy for %a@." pp_strategy_info config;
  let spec = Kernel_function.get_spec config.kf in
  let annots = Hannots.create 7 in
  let has_inv_prop = ref [] in
  let add_on_edges new_acc edges = 
    if new_acc.p_inv <> [] then has_inv_prop := edges @ !has_inv_prop;
    add_on_edges annots new_acc edges 
  in
  let add_loop_annots v s =
     let edges_before = Cil2cfg.pred_e cfg v in
     let edges_to_head = Cil2cfg.succ_e cfg v in
     let (loop_entry , loop_back , loop_core) =
       get_loop_annots config edges_before s
     in
     let back_edges, entry_edges =
       List.partition Cil2cfg.is_back_edge edges_before
     in
       begin
         add_on_edges loop_back back_edges;
         add_on_edges loop_entry entry_edges;
         add_on_edges loop_core edges_to_head;
       end
  in

  let edges_before v = Cil2cfg.pred_e cfg v in
  let edges_after v =
    try let v' = Cil2cfg.node_after cfg v in Cil2cfg.pred_e cfg v'
    with Not_found -> []
  in
  let edges_exits v = Cil2cfg.get_exit_edges cfg v in

  let add_annots v (before_annots, (post_annots, exits_annots)) =

    add_on_edges before_annots (edges_before v);

      if post_annots <> empty_acc then
        begin
          let edges_after = edges_after v in
          if edges_after = [] then
            Wp_parameters.warning ~once:true
              "Ignoring annotation rooted after statement with no succ"
          else add_on_edges post_annots edges_after
        end;

      if exits_annots <> empty_acc then
        begin
          let edges_exits = edges_exits v in
            if edges_exits = [] then
              List.iter (fun (pid,_) -> set_unreachable pid)
                exits_annots.p_goal
            else add_on_edges exits_annots edges_exits

        end
  in
  let add_stmt_annots ?(loop=false) v s =
    let l_post = match edges_after v with [] -> None
      | e::_ ->
          match Cil2cfg.get_edge_next_stmt cfg e with
            | None -> None
            | Some s ->  Some (Clabels.mk_logic_label s)
    in
    let stmt_annots = get_stmt_annots config ~loop (edges_before v) s l_post in
      add_annots v stmt_annots ;

      if loop then add_loop_annots v s
  in
  let add_fct_pre_annots v =
    let edges = Cil2cfg.succ_e cfg v in
    let pre = add_fct_pre config empty_acc spec in
    let _ = add_on_edges pre edges in
    let lp = add_behaviors_props config Kglobal spec empty_acc in
    let _ = add_on_edges lp edges in
    let _ = get_variant spec in (* TODO *)
    let _ = get_terminates spec in (* TODO *)
      ()
  in
  let add_fct_post_annots tkind v =
    let edges = Cil2cfg.succ_e cfg v in
    let post = get_fct_post config tkind spec in add_on_edges post edges;
    let assigns_kind = match tkind with
      | Normal -> if Cil2cfg.has_exit cfg then Assigns_FctOut else Assigns_Stmt
      | Exits -> Assigns_FctExit
      | _ -> assert false
    in
    let assigns = get_fct_assigns config assigns_kind spec in
      add_on_edges assigns edges
  in
  let get_node_annot v =
    let l_post = match edges_after v with [] -> None
      | e::_ ->
          match Cil2cfg.get_edge_next_stmt cfg e with
            | None -> None
            | Some s ->  Some (Clabels.mk_logic_label s)
    in
    Wp_parameters.debug ~level:2 "get_node_annot for node %a" Cil2cfg.pp_node v;
    match Cil2cfg.node_type v with
    | Cil2cfg.Vstart | Cil2cfg.Vend -> ()
    | Cil2cfg.VfctIn -> add_fct_pre_annots v
    | Cil2cfg.VfctOut  -> add_fct_post_annots Normal v
    | Cil2cfg.Vexit -> add_fct_post_annots Exits v

    | Cil2cfg.VblkIn (Cil2cfg.Bstmt s, _)
    | Cil2cfg.Vstmt s 
    | Cil2cfg.Vswitch (s,_) | Cil2cfg.Vtest (true, s, _)
      -> add_stmt_annots v s

    | Cil2cfg.Vcall (s,_,fct,_) -> 
      add_stmt_annots v s; 
      let call_annots = get_call_annots config s fct l_post in
      add_annots v call_annots 

    | Cil2cfg.Vloop (_, s) ->
        add_stmt_annots ~loop:true v s

    | Cil2cfg.VblkIn (_, _) | Cil2cfg.VblkOut (_, _) -> (* nothing *) ()
    | Cil2cfg.Vtest (false, _s, _) -> (* done in Cil2cfg.Vtest (true) *) ()
  in
    Cil2cfg.iter_nodes get_node_annot cfg;
    !has_inv_prop, annots

(* ------------------------------------------------------------------------ *)
(* ---  Global Properties                                               --- *)
(* ------------------------------------------------------------------------ *)

let get_global_annotations () =
  let rec do_g acc g =
    let source = Cil.source (Cil_datatype.Global_annotation.loc g) in
    match g with
    | Daxiomatic (_ax_name, globs,_) -> do_globs globs acc
    | Dfun_or_pred _ ->
	(* will be processed while translation is needed *) acc
    | Dtype _ ->
	(* will be processed while translation is needed *) acc
    | Dtype_annot (linfo,_) ->
        Wp_parameters.warning ~source
	  "Type invariant not handled yet ('%s' ignored)"
          linfo.l_var_info.lv_name;
        acc
    | Dinvariant (linfo,_) ->
        Wp_parameters.warning ~source
          "Global invariant not handled yet ('%s' ignored)"
          linfo.l_var_info.lv_name;
        acc
    | Dlemma (name, is_axiom, labels, _, pred,_) ->
	if not is_axiom then
          Wp_parameters.warning ~once:true ~source
            "Proof obligation for property '%s' not generated." name ;
	add_glob_axiom name acc labels pred
  and do_globs globs acc =
    List.fold_left do_g acc globs
  in
  let globs = Globals.Annotations.get_all () in
  let globs = List.map (fun (g, _generated) -> g) globs in
  let axioms = do_globs globs [] in
  axioms

(* ------------------------------------------------------------------------ *)
(* --- Main functions to build the strategies                           --- *)
(* ------------------------------------------------------------------------ *)

let behavior_name_of_config config =
  match config.cur_bhv with
    | FunBhv b when b = Cil.default_behavior_name -> None
    | FunBhv b -> Some b
    | StmtBhv (_, s, b) when b = Cil.default_behavior_name ->
        Some ("default_for_stmt_"^(string_of_int s.sid))(*TODO better name ?*)
    | StmtBhv (_, s, b) -> Some (b^"_stmt_"^(string_of_int s.sid))

let build_bhv_strategy config =
  let cfg = config.cfg in
  has_asgn_goal := false;
  has_prop_goal := false;
  let has_inv_prop, annots = get_behavior_annots cfg config in
  {
    behavior_name = behavior_name_of_config config;
    info = config;

    has_asgn_goal = !has_asgn_goal;
    has_prop_goal = !has_prop_goal;
    has_inv_prop = has_inv_prop;

    get_annots =
      (fun e -> try Hannots.find annots e with Not_found -> empty_acc);

    global_axioms = get_global_annotations ();

    strategy_kind = SKannots;
  }

(* Visit the CFG to find all the internal statement specifications.
 * (see [HdefAnnotBhv] documentation for infomation about this table).
 *)
let internal_function_behaviors cfg =
  let def_annot_bhv = HdefAnnotBhv.create 42 in
  let get_stmt_bhv node stmt acc =
    let add_bhv_info acc b =
      if b.b_name = Cil.default_behavior_name then
        begin
          let _, int_edges = Cil2cfg.get_internal_edges cfg node in
          let n = Cil2cfg.Eset.cardinal int_edges in
          let reg e =
            try
              let (_old_s, old_n) = HdefAnnotBhv.find def_annot_bhv e in
                if n < old_n then
                  (* new spec is included in the old one : override. *)
                  raise Not_found
            with Not_found ->
              HdefAnnotBhv.replace def_annot_bhv e (stmt, n)
          in
            Cil2cfg.Eset.iter reg int_edges
        end;
      (node, stmt, b.b_name)::acc
    in
    let spec_bhv_names acc annot = match annot with
      | {annot_content = AStmtSpec spec} ->
          List.fold_left add_bhv_info acc spec.spec_behavior
      | _ -> Wp_parameters.fatal "filter on is_contract didn't work ?"
    in
    let annots = Annotations.get_filter Logic_utils.is_contract stmt in
    let annots = List.map Annotations.get_code_annotation annots in
      List.fold_left spec_bhv_names acc annots
  in
  let get_bhv n ((seen_stmts, bhvs) as l) =
    match Cil2cfg.start_stmt_of_node n with None -> l
    | Some s ->
        if List.mem s.sid seen_stmts then l
        else
          let seen_stmts = s.sid::seen_stmts in
          let bhvs = get_stmt_bhv n s bhvs in
            (seen_stmts, bhvs)
  in
  let _, bhvs = Cil2cfg.fold_nodes get_bhv cfg ([], []) in
    bhvs, def_annot_bhv

let get_asked_bhv_from_name s_bhvs f_bhvs ki bname =
  let add_fct_bhv acc b =
    let add = match b with
      | None   -> bname = Cil.default_behavior_name
      | Some b -> b = bname && b <> Cil.default_behavior_name
    in
    match ki with
      | None (* not specified ki *)
      | Some Kglobal ->
          if add then (FunBhv bname)::acc else acc
      | Some Kstmt _ -> acc
  in

  let add_stmt_bhv acc (n,s,b) =
    if b = bname then
      match ki with
        | None -> (* not specified ki *) (StmtBhv (n, s, b))::acc
        | Some (Kstmt stmt) when stmt.sid = s.sid ->
            (StmtBhv (n, s, b))::acc
        | _ -> (* specified ki but not this one *) acc
      else acc
  in
  let acc = [] in
  let acc = add_fct_bhv acc None in (* default fct bhv *)
  let acc = List.fold_left add_stmt_bhv acc s_bhvs in
  let acc = List.fold_left (fun acc b -> add_fct_bhv acc (Some b)) acc f_bhvs in
    acc

let process_unreached_annots cfg unreached =
  Wp_parameters.debug ~level:3 "[strategy] collecting unreachable annotations@.";
  let kf = Cil2cfg.cfg_kf cfg in
  let spec = Kernel_function.get_spec kf in
  let add_id id acc =
    if filter_status id then id::acc
    else (* non-selected property : nothing to do *) acc
  in
  let do_post b tk acc (termk, _ as p) =
    if tk = termk then add_id (mk_fct_post_id kf b p) acc else acc
  in
  let do_bhv termk acc b = List.fold_left (do_post b termk) acc b.b_post_cond in
  let do_annot s a acc = match a with
    | Before (User ca | AI (_, ca)) -> add_id (mk_code_annot_id kf s ca) acc
    | _ -> (* TODO ? *) acc
  in
  let do_node acc n =
     Wp_parameters.debug ~level:3
       "[strategy] process annotations of unreachable node %a@."
       Cil2cfg.pp_node_type n;
    match n with
    | Cil2cfg.Vstart -> Wp_parameters.fatal "Start must be reachable"
    | Cil2cfg.VfctIn -> Wp_parameters.fatal "FctIn must be reachable"
    | Cil2cfg.VfctOut  -> List.fold_left (do_bhv Normal) acc spec.spec_behavior
    | Cil2cfg.Vexit  -> List.fold_left (do_bhv Exits) acc spec.spec_behavior
    | Cil2cfg.Vstmt s
    | Cil2cfg.VblkIn (Cil2cfg.Bstmt s, _) | Cil2cfg.Vcall (s, _, _, _)
    | Cil2cfg.Vtest (true, s, _) | Cil2cfg.Vloop (_, s) | Cil2cfg.Vswitch (s,_)
      -> Annotations.single_fold_stmt (do_annot s) s acc
    | Cil2cfg.Vtest (false, _, _)
    | Cil2cfg.VblkIn _ | Cil2cfg.VblkOut _ | Cil2cfg.Vend -> acc
  in
  let annots = List.fold_left do_node [] unreached in
    Wp_parameters.debug ~level:3
      "[strategy] found %d unreachable annotations@." (List.length annots) ;
  List.iter (fun pid -> set_unreachable pid) annots

(*----------------------------------------------------------------------------*)
(* Specific function to check functional dependencies                         *)
(*----------------------------------------------------------------------------*)

let get_loop_assigns_for_from config s =
  let do_ca ca acc = 
    match ca.annot_content with
      | AAssigns (b_list, Writes a) 
          when 
            List.exists (fun x -> x = name_of_asked_bhv config.cur_bhv) b_list
            -> Some (ca,a)
      | _ -> acc
  in
  let do_annot a acc = match a with
    | Before (User ca | AI (_, ca)) -> do_ca ca acc
    | After _ -> acc
  in Annotations.single_fold_stmt do_annot s None

let mk_assign_annots_for_froms config bhv _fct_pre assign =
  let annots = Hannots.create 7 in
  let add_post v p = match p with None -> ()
    | Some post ->
        let edges = Cil2cfg.succ_e config.cfg v in
        let assumes = Logic_const.ptrue in (* assumes used in pre. *)
        let post = prepare_fct_post config bhv Normal assumes post in
        let post = add_prop config empty_acc Agoal post in
          add_on_edges annots post edges
  in
  let get_node_annot v = match Cil2cfg.node_type v with
    | Cil2cfg.VfctIn -> ()
        (* Don't put the precondition here because we don't want to build
        * (pre => post) => (pre' => post') but rather 
        * (pre /\ pre' /\ post => post') so we have to process the pre latter
        * (see SKfroms) *)

    | Cil2cfg.VfctOut  -> add_post v (WpFroms.mk_assign_post Normal assign)
    | Cil2cfg.Vexit -> add_post v (WpFroms.mk_assign_post Exits assign)

    | Cil2cfg.VblkIn (Cil2cfg.Bstmt _s, _)
    | Cil2cfg.Vstmt _s 
    | Cil2cfg.Vswitch (_s,_) | Cil2cfg.Vtest (true, _s, _)
      -> (* TODO: add_stmt_annots v s *) ()

    | Cil2cfg.Vcall (_s,_,_fct,_) ->
        (* TODO: but have to reorganize to be able to call [add_annots]...
        let call_annots = get_call_annots config s fct in
          add_annots v call_annots 
          *) ()

    | Cil2cfg.Vloop (_, s) ->
        (* TODO: add_stmt_annots ~loop:true v s *) 
        (* at least need to give the loop assigns to compute something correct
         * ([] when no user loop assigns).
         * TODO: this should be changed because the strategy shouldn't be able
         * to produce an incorrect computation ! *)
          (*
     let edges_before = Cil2cfg.pred_e cfg v in
     let back_edges, entry_edges =
       List.partition Cil2cfg.is_back_edge edges_before
     in
       *)
     let edges_after = Cil2cfg.succ_e config.cfg v in
     (match get_loop_assigns_for_from config s with
         None -> ()
       | Some assigns ->
         let p_assign = add_loop_assigns config Ahyp s assigns empty_acc in
         add_on_edges annots p_assign edges_after)
    | _ -> ()
  in
  let _ = Cil2cfg.iter_nodes get_node_annot config.cfg in
    annots


let mk_assign_strategy_for_froms config bhv pre n (_,from as assign) =
  let bname = behavior_name_of_config config in
  let bname = match bname with None -> "default" | Some bname -> bname in
  let bname = (bname^"_assign_"^(string_of_int n)) in
  let get_pre () = 
    let pre_init = WpFroms.mk_assign_pre from in
    let pre_init = prepare_pre config Kglobal bhv pre_init in
    let pre = add_prop config pre Ahyp pre_init in
      pre
  in
  let annots = mk_assign_annots_for_froms config bhv pre assign in
  {
    behavior_name = Some bname;
    info = config;
    has_asgn_goal = false;
    has_prop_goal = true;
    has_inv_prop = [];
    get_annots =
      (fun e -> try Hannots.find annots e with Not_found -> empty_acc);

    global_axioms = get_global_annotations ();

    strategy_kind = SKfroms { get_pre = get_pre };
  }

let build_strategy_for_froms config =
  let spec = Kernel_function.get_spec config.kf in
  match config.cur_bhv with 
    | StmtBhv (_n, _s, _b) -> 
        Wp_parameters.warning "Not implemented: froms in stmt bhv"; []
    | FunBhv b -> 
        let bhv = get_behav config Kglobal spec.spec_behavior in
          match bhv with
            | None -> 
                if b <> Cil.default_behavior_name then
                Wp_parameters.warning 
                  "function behavior '%s' not found" b; []
            | Some bhv ->
                let pre = add_fct_pre config empty_acc spec in
                  (* precond of the function + assumes of the behavior *)
                let add_assign_strategy (n, acc as res) (b,f) =
                  match f with
                      FromAny -> res
                    | From l ->
                      let stg = 
                        mk_assign_strategy_for_froms config bhv pre n (b,l)
                      in 
                      (n+1, stg::acc)
                in
                match bhv.b_assigns with
                    WritesAny -> []
                  | Writes l ->
                    let _, acc = 
                      List.fold_left add_assign_strategy (1, []) l
                    in
                    acc

(*----------------------------------------------------------------------------*)
(* Everything must go through here.                                           *)
(*----------------------------------------------------------------------------*)

let do_froms () =  false (* Wp_parameters.Froms.get () *)

let get_strategies assigns cfg behaviors ki property =
  Wp_parameters.feedback ~level:3
    "[wpAnnot.get_strategies] for behaviors names: %a@."
    (Wp_error.pp_string_list ~sep:" " ~empty:"<none>")
    (match behaviors with None -> ["<all>"] | Some l -> l) ;
  let _ = match ki with
    | None -> ()
    | Some Kglobal ->
        Wp_parameters.feedback ~level:3
          "[wpAnnot.get_strategies] select in function properies@."
    | Some (Kstmt s) ->
        Wp_parameters.feedback ~level:3
          "[wpAnnot.get_strategies] select stmt %d properties@." s.sid
  in
  let kf = Cil2cfg.cfg_kf cfg in
  let s_bhvs, def_annot_bhv = internal_function_behaviors cfg in
  let bhvs =
    let f_bhvs = Kernel_function.spec_function_behaviors kf in
    match behaviors with
      | Some bhvs ->
          List.fold_left
            (fun acc b -> (get_asked_bhv_from_name s_bhvs f_bhvs ki b)@acc)
            [] bhvs
      | None ->
          let spec = Kernel_function.get_spec kf in
          (* default behavior name is always added even if it doesn't appear
            * explicitly in spec in order to process internal properties. *)
          let bhvs = get_behavior_names ~with_default:true spec in
          let bhvs = List.map (fun b -> FunBhv b) bhvs in
            List.fold_left
              (fun acc (n,s,b) -> (StmtBhv (n,s,b))::acc) bhvs s_bhvs
  in
    if bhvs = [] then
      Wp_parameters.warning "[wpAnnot.get_strategies] no behaviors found";
  let do_bhv bhv =
    let config which = { kf = Cil2cfg.cfg_kf cfg ;
                   cfg = cfg;
                   which_strategy = which;
                   new_loops = new_loops;
                   cur_bhv = bhv;
                   asked_prop = property;
                   asked_bhvs = bhvs;
                   assigns_filter = assigns;
                   def_annots_info = def_annot_bhv}
    in if do_froms () 
    then build_strategy_for_froms (config WSfroms)
    else [build_bhv_strategy (config WSannots)]
  in List.flatten (List.map do_bhv bhvs)

(*----------------------------------------------------------------------------*)
(* Public functions to build the strategies                                   *)
(*----------------------------------------------------------------------------*)

let get_cfg ~with_rte kf =
  if with_rte then compute_rte_for kf ;
  let cfg, unreached = Cil2cfg.get kf in
    process_unreached_annots cfg unreached ;
    cfg

(* TODO: it could be better to first find the property in order
 * to compute only in its behaviors...
 * At the moment [p_bhvs] is given as a hint (see [get_id_prop_strategies])
 *)
let get_prop_strategies ?(assigns=WithAssigns) cfg (p_bhvs, prop_name) =
  get_strategies assigns cfg p_bhvs None (NamedProp prop_name)

let get_id_prop_strategies ?(assigns=WithAssigns) cfg p =
  Wp_parameters.debug ~level:3 "[wpAnnot.get_id_prop_strategies] %s@."
    (id_prop_txt p);
  let all_behaviors = None in
  let default_behavior = Some [ Cil.default_behavior_name ] in
  let get_code_annot_bhv annot =
    match annot.annot_content with
      | AAssert (l, _) | AInvariant (l, _, _) | AAssigns (l, _) ->
          ( match l with [] -> all_behaviors | _ -> Some l )
      | _ -> all_behaviors
  in
  let b_of_fb_opt fb_opt = match fb_opt with
    | None -> default_behavior
    | Some fb -> Some [fb.b_name]
  in
  let ki, bhvs =
    match p with
      | Property.IPCodeAnnot (_,_,ca) ->
          None, get_code_annot_bhv ca
      | Property.IPAssigns (_, _, Property.Id_code_annot _, _) -> 
          (*loop assigns *)
          None, default_behavior
      | Property.IPPredicate (Property.PKRequires _, _kf, Kglobal, _p)
        -> assert false (* get_precond_strategies should have been called !*)
      | _ ->
          Some (Property.get_kinstr p),
          b_of_fb_opt (Property.get_behavior p)
  in
  get_strategies assigns cfg bhvs ki (IdProp p)

(*TODO[LC] removed unless better implementation is found.
  WARNING: invoking strategies in the GUI should never raise uncaught exception.
  SEE: bug on Frama-C GUI Exit

let get_precond_strategies p =
  Wp_parameters.debug ~level:3 "[wpAnnot.get_precond_strategies] %s@."
    (id_prop_txt p);
  match p with
    | Property.IPPredicate (Property.PKRequires _, kf, Kglobal, _) ->
        if is_main_entry_point kf then
          let cfg = get_cfg ~with_rte:false kf in
          (* Wp_parameters.not_yet_implemented "check the [main] precond" *)
          get_strategies NoAssigns cfg None None (IdProp p)
        else if Db.Value.is_computed () then
          let idp = (IdProp p) in
          let strategies = ref [] in
          let do_caller (caller:kernel_function) =
            let cfg = get_cfg ~with_rte:false caller in
            let caller_annots = get_strategies NoAssigns cfg None None idp in
              strategies := caller_annots @ !strategies
          in
          !Db.Semantic_Callgraph.iter_on_callers do_caller kf;
          !strategies
        else
          Wp_parameters.not_yet_implemented
            "Find callers of %a without value analysis"
             Kernel_function.pretty_name kf
    | _ -> 
        raise (Invalid_argument 
                 "[get_precond_strategies] not a function precondition")
*)

let get_behavior_strategies ?(assigns=WithAssigns) cfg bhv =
  get_strategies assigns cfg (Some bhv) None AllProps

let get_function_strategies ?(assigns=WithAssigns) cfg =
  get_strategies assigns cfg None None AllProps

(*----------------------------------------------------------------------------*)
(*
Local Variables:
compile-command: "make -C ../.."
End:
*)
(*----------------------------------------------------------------------------*)
